import asyncio
import contextlib
import os
import random
import sys
from unittest.mock import Mock

import pytest
from sanic import Sanic, request, response
from sanic import __version__ as SANIC_VERSION_RAW
from sanic.exceptions import SanicException
from sanic.response import HTTPResponse

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.integrations.sanic import SanicIntegration
from sentry_sdk.tracing import TransactionSource
from tests.conftest import get_free_port

try:
    from sanic_testing import TestManager
except ImportError:
    TestManager = None

try:
    from sanic_testing.reusable import ReusableClient
except ImportError:
    ReusableClient = None

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Container, Iterable
    from typing import Any, Optional

SANIC_VERSION = tuple(map(int, SANIC_VERSION_RAW.split(".")))
PERFORMANCE_SUPPORTED = SANIC_VERSION >= (21, 9)


@pytest.fixture
def app():
    if SANIC_VERSION < (19,):
        """
        Older Sanic versions 0.8 and 18 bind to the same fixed port which
        creates problems when we run tests concurrently.
        """
        old_test_client = Sanic.test_client.__get__

        def new_test_client(self):
            client = old_test_client(self, Sanic)
            client.port += os.getpid() % 100
            return client

        Sanic.test_client = property(new_test_client)

    if SANIC_VERSION >= (20, 12) and SANIC_VERSION < (22, 6):
        # Some builds (20.12.0 intruduced and 22.6.0 removed again) have a feature where the instance is stored in an internal class
        # registry for later retrieval, and so add register=False to disable that
        sanic_app = Sanic("Test", register=False)
    else:
        sanic_app = Sanic("Test")

    if TestManager is not None:
        TestManager(sanic_app)

    @sanic_app.route("/message")
    def hi(request):
        capture_message("hi")
        return response.text("ok")

    @sanic_app.route("/message/<message_id>")
    def hi_with_id(request, message_id):
        capture_message("hi with id")
        return response.text("ok with id")

    @sanic_app.route("/500")
    def fivehundred(_):
        1 / 0

    return sanic_app


def get_client(app):
    @contextlib.contextmanager
    def simple_client(app):
        yield app.test_client

    if ReusableClient is not None:
        return ReusableClient(app, port=get_free_port())
    else:
        return simple_client(app)


def test_request_data(sentry_init, app, capture_events):
    sentry_init(integrations=[SanicIntegration()])
    events = capture_events()

    c = get_client(app)
    with c as client:
        _, response = client.get("/message?foo=bar")
        assert response.status == 200

    (event,) = events
    assert event["transaction"] == "hi"
    assert event["request"]["env"] == {"REMOTE_ADDR": ""}
    assert set(event["request"]["headers"]) >= {
        "accept",
        "accept-encoding",
        "host",
        "user-agent",
    }
    assert event["request"]["query_string"] == "foo=bar"
    assert event["request"]["url"].endswith("/message")
    assert event["request"]["method"] == "GET"

    # Assert that state is not leaked
    events.clear()
    capture_message("foo")
    (event,) = events

    assert "request" not in event
    assert "transaction" not in event


@pytest.mark.parametrize(
    "url,expected_transaction,expected_source",
    [
        ("/message", "hi", "component"),
        ("/message/123456", "hi_with_id", "component"),
    ],
)
def test_transaction_name(
    sentry_init, app, capture_events, url, expected_transaction, expected_source
):
    sentry_init(integrations=[SanicIntegration()])
    events = capture_events()

    c = get_client(app)
    with c as client:
        _, response = client.get(url)
        assert response.status == 200

    (event,) = events
    assert event["transaction"] == expected_transaction
    assert event["transaction_info"] == {"source": expected_source}


def test_errors(sentry_init, app, capture_events):
    sentry_init(integrations=[SanicIntegration()])
    events = capture_events()

    @app.route("/error")
    def myerror(request):
        raise ValueError("oh no")

    c = get_client(app)
    with c as client:
        _, response = client.get("/error")
        assert response.status == 500

    (event,) = events
    assert event["transaction"] == "myerror"
    (exception,) = event["exception"]["values"]

    assert exception["type"] == "ValueError"
    assert exception["value"] == "oh no"
    assert any(
        frame["filename"].endswith("test_sanic.py")
        for frame in exception["stacktrace"]["frames"]
    )


def test_bad_request_not_captured(sentry_init, app, capture_events):
    sentry_init(integrations=[SanicIntegration()])
    events = capture_events()

    @app.route("/")
    def index(request):
        raise SanicException("...", status_code=400)

    c = get_client(app)
    with c as client:
        _, response = client.get("/")
        assert response.status == 400

    assert not events


def test_error_in_errorhandler(sentry_init, app, capture_events):
    sentry_init(integrations=[SanicIntegration()])
    events = capture_events()

    @app.route("/error")
    def myerror(request):
        raise ValueError("oh no")

    @app.exception(ValueError)
    def myhandler(request, exception):
        1 / 0

    c = get_client(app)
    with c as client:
        _, response = client.get("/error")
        assert response.status == 500

    event1, event2 = events

    (exception,) = event1["exception"]["values"]
    assert exception["type"] == "ValueError"
    assert any(
        frame["filename"].endswith("test_sanic.py")
        for frame in exception["stacktrace"]["frames"]
    )

    exception = event2["exception"]["values"][-1]
    assert exception["type"] == "ZeroDivisionError"
    assert any(
        frame["filename"].endswith("test_sanic.py")
        for frame in exception["stacktrace"]["frames"]
    )


def test_concurrency(sentry_init, app):
    """
    Make sure we instrument Sanic in a way where request data does not leak
    between request handlers. This test also implicitly tests our concept of
    how async code should be instrumented, so if it breaks it likely has
    ramifications for other async integrations and async usercode.

    We directly call the request handler instead of using Sanic's test client
    because that's the only way we could reproduce leakage with such a low
    amount of concurrent tasks.
    """
    sentry_init(integrations=[SanicIntegration()])

    @app.route("/context-check/<i>")
    async def context_check(request, i):
        scope = sentry_sdk.get_isolation_scope()
        scope.set_tag("i", i)

        await asyncio.sleep(random.random())

        scope = sentry_sdk.get_isolation_scope()
        assert scope._tags["i"] == i

        return response.text("ok")

    async def task(i):
        responses = []

        kwargs = {
            "url_bytes": "http://localhost/context-check/{i}".format(i=i).encode(
                "ascii"
            ),
            "headers": {},
            "version": "1.1",
            "method": "GET",
            "transport": None,
        }

        if SANIC_VERSION >= (19,):
            kwargs["app"] = app

        if SANIC_VERSION >= (21, 3):

            class MockAsyncStreamer:
                def __init__(self, request_body):
                    self.request_body = request_body
                    self.iter = iter(self.request_body)

                    if SANIC_VERSION >= (21, 12):
                        self.response = None
                        self.stage = Mock()
                    else:
                        self.response = b"success"

                def respond(self, response):
                    responses.append(response)
                    patched_response = HTTPResponse()
                    return patched_response

                def __aiter__(self):
                    return self

                async def __anext__(self):
                    try:
                        return next(self.iter)
                    except StopIteration:
                        raise StopAsyncIteration

            patched_request = request.Request(**kwargs)
            patched_request.stream = MockAsyncStreamer([b"hello", b"foo"])

            if SANIC_VERSION >= (21, 9):
                await app.dispatch(
                    "http.lifecycle.request",
                    context={"request": patched_request},
                    inline=True,
                )

            await app.handle_request(
                patched_request,
            )
        else:
            await app.handle_request(
                request.Request(**kwargs),
                write_callback=responses.append,
                stream_callback=responses.append,
            )

        (r,) = responses
        assert r.status == 200

    async def runner():
        if SANIC_VERSION >= (21, 3):
            if SANIC_VERSION >= (21, 9):
                await app._startup()
            else:
                try:
                    app.router.reset()
                    app.router.finalize()
                except AttributeError:
                    ...
        await asyncio.gather(*(task(i) for i in range(1000)))

    if sys.version_info < (3, 7):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(runner())
    else:
        asyncio.run(runner())

    scope = sentry_sdk.get_isolation_scope()
    assert not scope._tags


class TransactionTestConfig:
    """
    Data class to store configurations for each performance transaction test run, including
    both the inputs and relevant expected results.
    """

    def __init__(
        self,
        integration_args: "Iterable[Optional[Container[int]]]",
        url: str,
        expected_status: int,
        expected_transaction_name: "Optional[str]",
        expected_source: "Optional[str]" = None,
        streaming_compatible: bool = True,
    ) -> None:
        """
        expected_transaction_name of None indicates we expect to not receive a transaction
        """
        self.integration_args = integration_args
        self.url = url
        self.expected_status = expected_status
        self.expected_transaction_name = expected_transaction_name
        self.expected_source = expected_source
        self.streaming_compatible = streaming_compatible


@pytest.mark.skipif(
    not PERFORMANCE_SUPPORTED, reason="Performance not supported on this Sanic version"
)
@pytest.mark.parametrize("send_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize(
    "test_config",
    [
        TransactionTestConfig(
            # Transaction for successful page load
            integration_args=(),
            url="/message",
            expected_status=200,
            expected_transaction_name="hi",
            expected_source=TransactionSource.COMPONENT,
        ),
        TransactionTestConfig(
            # Transaction for successful page load with query string
            integration_args=(),
            url="/message?foo=bar",
            expected_status=200,
            expected_transaction_name="hi",
            expected_source=TransactionSource.COMPONENT,
        ),
        TransactionTestConfig(
            # Transaction still recorded when we have an internal server error
            integration_args=(),
            url="/500",
            expected_status=500,
            expected_transaction_name="fivehundred",
            expected_source=TransactionSource.COMPONENT,
        ),
        TransactionTestConfig(
            # By default, no transaction when we have a 404 error
            integration_args=(),
            url="/404",
            expected_status=404,
            expected_transaction_name=None,
            streaming_compatible=False,
        ),
        TransactionTestConfig(
            # With no ignored HTTP statuses, we should get transactions for 404 errors
            integration_args=(None,),
            url="/404",
            expected_status=404,
            expected_transaction_name="/404",
            expected_source=TransactionSource.URL,
        ),
        TransactionTestConfig(
            # Transaction can be suppressed for other HTTP statuses, too, by passing config to the integration
            integration_args=({200},),
            url="/message",
            expected_status=200,
            expected_transaction_name=None,
            streaming_compatible=False,
        ),
    ],
)
def test_transactions(
    test_config: "TransactionTestConfig",
    sentry_init: "Any",
    app: "Any",
    capture_events: "Any",
    capture_items: "Any",
    span_streaming: bool,
    send_pii: bool,
) -> None:
    if span_streaming and not test_config.streaming_compatible:
        pytest.skip("unsampled_statuses is not supported in span streaming mode")

    # Init the SanicIntegration with the desired arguments
    sentry_init(
        integrations=[SanicIntegration(*test_config.integration_args)],
        traces_sample_rate=1.0,
        send_default_pii=send_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    # Make request to the desired URL
    c = get_client(app)
    with c as client:
        _, response = client.get(test_config.url)
        assert response.status == test_config.expected_status

    sentry_sdk.flush()

    if span_streaming:
        segments = [
            i.payload
            for i in items
            if i.payload["attributes"].get("sentry.origin") == "auto.http.sanic"
            and i.payload["is_segment"]
        ]
        assert len(segments) <= 1
        (segment, *_) = [*segments, None]

        assert (segment is None) == (test_config.expected_transaction_name is None)

        if segment is not None:
            assert segment["name"] == test_config.expected_transaction_name
            assert (
                segment["attributes"]["sentry.segment.name.source"]
                == test_config.expected_source
            )

            attrs = segment["attributes"]
            assert attrs["http.request.method"] == "GET"
            assert attrs["network.protocol.name"] == "http"
            header_keys = {
                key[len("http.request.header.") :]
                for key in attrs
                if key.startswith("http.request.header.")
            }
            assert header_keys >= {"accept", "accept-encoding", "host", "user-agent"}
            assert attrs["http.response.status_code"] == test_config.expected_status
            assert segment["status"] == (
                "error" if test_config.expected_status >= 400 else "ok"
            )

            if send_pii:
                assert attrs["url.full"].endswith(test_config.url)
                assert attrs["url.path"] == test_config.url.split("?")[0]
                if "?" in test_config.url:
                    assert attrs["http.query"] == test_config.url.split("?", 1)[1]

            else:
                assert "url.full" not in attrs
                assert "url.path" not in attrs
                assert "http.query" not in attrs

    else:
        # Extract the transaction events by inspecting the event types. We should at most have 1 transaction event.
        transaction_events = [
            e for e in events if "type" in e and e["type"] == "transaction"
        ]
        assert len(transaction_events) <= 1

        # Get the only transaction event, or set to None if there are no transaction events.
        (transaction_event, *_) = [*transaction_events, None]

        # We should have no transaction event if and only if we expect no transactions
        assert (transaction_event is None) == (
            test_config.expected_transaction_name is None
        )

        # If a transaction was expected, ensure it is correct
        assert (
            transaction_event is None
            or transaction_event["transaction"] == test_config.expected_transaction_name
        )
        assert (
            transaction_event is None
            or transaction_event["transaction_info"]["source"]
            == test_config.expected_source
        )


@pytest.mark.skipif(
    not PERFORMANCE_SUPPORTED, reason="Performance not supported on this Sanic version"
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(sentry_init, app, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[SanicIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    c = get_client(app)
    with c as client:
        client.get("/message?foo=bar")

    sentry_sdk.flush()

    if span_streaming:
        (segment,) = [
            i.payload
            for i in items
            if i.payload["attributes"].get("sentry.origin") == "auto.http.sanic"
        ]
        assert segment["attributes"]["sentry.origin"] == "auto.http.sanic"
    else:
        (_, event) = events
        assert event["contexts"]["trace"]["origin"] == "auto.http.sanic"


@pytest.mark.skipif(
    not PERFORMANCE_SUPPORTED, reason="Performance not supported on this Sanic version"
)
@pytest.mark.parametrize("send_default_pii", [True, False])
def test_user_ip_address_on_all_spans(
    sentry_init, app, capture_items, send_default_pii
):
    app.config.FORWARDED_SECRET = "test"

    @app.route("/child-span")
    def child_span_handler(request):
        with sentry_sdk.traces.start_span(name="child-span"):
            pass
        return response.text("ok")

    sentry_init(
        integrations=[SanicIntegration()],
        default_integrations=False,
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream",
    )

    items = capture_items("span")

    c = get_client(app)
    with c as client:
        client.get(
            "/child-span",
            headers={"Forwarded": "for=127.0.0.1;secret=test"},
        )

    sentry_sdk.flush()

    child_span, server_span = [item.payload for item in items]

    if send_default_pii:
        assert server_span["attributes"]["user.ip_address"] == "127.0.0.1"
        assert child_span["attributes"]["user.ip_address"] == "127.0.0.1"
    else:
        assert "user.ip_address" not in server_span["attributes"]
        assert "user.ip_address" not in child_span["attributes"]



import os
import signal
import tempfile
import threading
import time

from celery.beat import Scheduler

from sentry_sdk.utils import logger


class ImmediateScheduler(Scheduler):
    """
    A custom scheduler that starts tasks immediately after starting Celery beat.
    """

    def setup_schedule(self):
        super().setup_schedule()
        for _, entry in self.schedule.items():
            self.apply_entry(entry)

    def tick(self):
        # Override tick to prevent the normal schedule cycle
        return 1


def kill_beat(beat_pid_file, delay_seconds=1):
    """
    Terminates Celery Beat after the given `delay_seconds`.
    """
    logger.info("Starting Celery Beat killer...")
    time.sleep(delay_seconds)
    pid = int(open(beat_pid_file, "r").read())
    logger.info("Terminating Celery Beat...")
    os.kill(pid, signal.SIGTERM)


def run_beat(celery_app, runtime_seconds=1, loglevel="warning", quiet=True):
    """
    Run Celery Beat that immediately starts tasks.
    The Celery Beat instance is automatically terminated after `runtime_seconds`.
    """
    logger.info("Starting Celery Beat...")
    pid_file = os.path.join(tempfile.mkdtemp(), f"celery-beat-{os.getpid()}.pid")

    t = threading.Thread(
        target=kill_beat,
        args=(pid_file,),
        kwargs={"delay_seconds": runtime_seconds},
    )
    t.start()

    beat_instance = celery_app.Beat(
        loglevel=loglevel,
        quiet=quiet,
        pidfile=pid_file,
    )
    beat_instance.run()



import concurrent.futures as cf
import sys
from contextlib import contextmanager
from random import random
from unittest.mock import Mock

import pytest
from statsig import statsig
from statsig.statsig_user import StatsigUser

import sentry_sdk
from sentry_sdk import start_span, start_transaction
from sentry_sdk.integrations.statsig import StatsigIntegration
from tests.conftest import ApproxDict


@contextmanager
def mock_statsig(gate_dict):
    old_check_gate = statsig.check_gate

    def mock_check_gate(user, gate, *args, **kwargs):
        return gate_dict.get(gate, False)

    statsig.check_gate = Mock(side_effect=mock_check_gate)

    yield

    statsig.check_gate = old_check_gate


def test_check_gate(sentry_init, capture_events, uninstall_integration):
    uninstall_integration(StatsigIntegration.identifier)

    with mock_statsig({"hello": True, "world": False}):
        sentry_init(integrations=[StatsigIntegration()])
        events = capture_events()
        user = StatsigUser(user_id="user-id")

        statsig.check_gate(user, "hello")
        statsig.check_gate(user, "world")
        statsig.check_gate(user, "other")  # unknown gates default to False.

        sentry_sdk.capture_exception(Exception("something wrong!"))

        assert len(events) == 1
        assert events[0]["contexts"]["flags"] == {
            "values": [
                {"flag": "hello", "result": True},
                {"flag": "world", "result": False},
                {"flag": "other", "result": False},
            ]
        }


def test_check_gate_threaded(sentry_init, capture_events, uninstall_integration):
    uninstall_integration(StatsigIntegration.identifier)

    with mock_statsig({"hello": True, "world": False}):
        sentry_init(integrations=[StatsigIntegration()])
        events = capture_events()
        user = StatsigUser(user_id="user-id")

        # Capture an eval before we split isolation scopes.
        statsig.check_gate(user, "hello")

        def task(flag_key):
            # Creates a new isolation scope for the thread.
            # This means the evaluations in each task are captured separately.
            with sentry_sdk.isolation_scope():
                statsig.check_gate(user, flag_key)
                # use a tag to identify to identify events later on
                sentry_sdk.set_tag("task_id", flag_key)
                sentry_sdk.capture_exception(Exception("something wrong!"))

        with cf.ThreadPoolExecutor(max_workers=2) as pool:
            pool.map(task, ["world", "other"])

        # Capture error in original scope
        sentry_sdk.set_tag("task_id", "0")
        sentry_sdk.capture_exception(Exception("something wrong!"))

        assert len(events) == 3
        events.sort(key=lambda e: e["tags"]["task_id"])

        assert events[0]["contexts"]["flags"] == {
            "values": [
                {"flag": "hello", "result": True},
            ]
        }
        assert events[1]["contexts"]["flags"] == {
            "values": [
                {"flag": "hello", "result": True},
                {"flag": "other", "result": False},
            ]
        }
        assert events[2]["contexts"]["flags"] == {
            "values": [
                {"flag": "hello", "result": True},
                {"flag": "world", "result": False},
            ]
        }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires python3.7 or higher")
def test_check_gate_asyncio(sentry_init, capture_events, uninstall_integration):
    asyncio = pytest.importorskip("asyncio")
    uninstall_integration(StatsigIntegration.identifier)

    with mock_statsig({"hello": True, "world": False}):
        sentry_init(integrations=[StatsigIntegration()])
        events = capture_events()
        user = StatsigUser(user_id="user-id")

        # Capture an eval before we split isolation scopes.
        statsig.check_gate(user, "hello")

        async def task(flag_key):
            with sentry_sdk.isolation_scope():
                statsig.check_gate(user, flag_key)
                # use a tag to identify to identify events later on
                sentry_sdk.set_tag("task_id", flag_key)
                sentry_sdk.capture_exception(Exception("something wrong!"))

        async def runner():
            return asyncio.gather(task("world"), task("other"))

        asyncio.run(runner())

        # Capture error in original scope
        sentry_sdk.set_tag("task_id", "0")
        sentry_sdk.capture_exception(Exception("something wrong!"))

        assert len(events) == 3
        events.sort(key=lambda e: e["tags"]["task_id"])

        assert events[0]["contexts"]["flags"] == {
            "values": [
                {"flag": "hello", "result": True},
            ]
        }
        assert events[1]["contexts"]["flags"] == {
            "values": [
                {"flag": "hello", "result": True},
                {"flag": "other", "result": False},
            ]
        }
        assert events[2]["contexts"]["flags"] == {
            "values": [
                {"flag": "hello", "result": True},
                {"flag": "world", "result": False},
            ]
        }


def test_wraps_original(sentry_init, uninstall_integration):
    uninstall_integration(StatsigIntegration.identifier)
    flag_value = random() < 0.5

    with mock_statsig(
        {"test-flag": flag_value}
    ):  # patches check_gate with a Mock object.
        mock_check_gate = statsig.check_gate
        sentry_init(integrations=[StatsigIntegration()])  # wraps check_gate.
        user = StatsigUser(user_id="user-id")

        res = statsig.check_gate(user, "test-flag", "extra-arg", kwarg=1)  # type: ignore[arg-type]

        assert res == flag_value
        assert mock_check_gate.call_args == (  # type: ignore[attr-defined]
            (user, "test-flag", "extra-arg"),
            {"kwarg": 1},
        )


def test_wrapper_attributes(sentry_init, uninstall_integration):
    uninstall_integration(StatsigIntegration.identifier)
    original_check_gate = statsig.check_gate
    sentry_init(integrations=[StatsigIntegration()])

    # Methods have not lost their qualified names after decoration.
    assert statsig.check_gate.__name__ == "check_gate"
    assert statsig.check_gate.__qualname__ == original_check_gate.__qualname__

    # Clean up
    statsig.check_gate = original_check_gate


@pytest.mark.parametrize(
    "span_streaming",
    [True, False],
)
def test_statsig_span_integration(
    sentry_init, capture_events, capture_items, uninstall_integration, span_streaming
):
    uninstall_integration(StatsigIntegration.identifier)

    with mock_statsig({"hello": True}):
        sentry_init(
            traces_sample_rate=1.0,
            integrations=[StatsigIntegration()],
            trace_lifecycle="stream" if span_streaming else "static",
        )
        user = StatsigUser(user_id="user-id")

        if span_streaming:
            items = capture_items("span")
            with sentry_sdk.traces.start_span(name="hi"):
                statsig.check_gate(user, "hello")
                statsig.check_gate(user, "world")

            sentry_sdk.flush()

            assert len(items) == 1
            span = items[0].payload
            assert span["attributes"]["flag.evaluation.hello"] is True
            assert span["attributes"]["flag.evaluation.world"] is False

        else:
            events = capture_events()
            with start_transaction(name="hi"):
                with start_span(op="foo", name="bar"):
                    statsig.check_gate(user, "hello")
                    statsig.check_gate(user, "world")

            (event,) = events
            assert event["spans"][0]["data"] == ApproxDict(
                {"flag.evaluation.hello": True, "flag.evaluation.world": False}
            )



import asyncio
import gzip
import io
import json
import os
import socket
import warnings
from collections import namedtuple
from contextlib import contextmanager
from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from unittest import mock
from urllib.parse import parse_qs, urlparse

try:
    import brotli
except ImportError:
    brotli = None

try:
    import jsonschema
except ImportError:
    jsonschema = None

import pytest

try:
    from pytest_localserver.http import WSGIServer
except ImportError:
    WSGIServer = None

try:
    from werkzeug.wrappers import Request, Response
except ImportError:
    Request = None
    Response = None

try:
    from starlette.testclient import TestClient
    # Catch RuntimeError to prevent the following exception in aws_lambda tests.
    # RuntimeError: The starlette.testclient module requires the httpx package to be installed.
except (ImportError, RuntimeError):
    TestClient = None

try:
    import gevent
except ImportError:
    gevent = None

try:
    import eventlet
except ImportError:
    eventlet = None

import sentry_sdk
import sentry_sdk.utils
from sentry_sdk.envelope import Envelope, parse_json
from sentry_sdk.integrations import (  # noqa: F401
    _DEFAULT_INTEGRATIONS,
    _installed_integrations,
    _processed_integrations,
)
from sentry_sdk.profiler import teardown_profiler
from sentry_sdk.profiler.continuous_profiler import teardown_continuous_profiler
from sentry_sdk.transport import Transport
from sentry_sdk.utils import package_version, reraise

try:
    import openai
except ImportError:
    openai = None


try:
    import anthropic
except ImportError:
    anthropic = None


try:
    import google
except ImportError:
    google = None


from typing import TYPE_CHECKING

from tests import _warning_recorder, _warning_recorder_mgr

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any, Callable, MutableMapping, Optional

try:
    from httpx import (
        ASGITransport,
        AsyncByteStream,
        AsyncClient,
    )
    from httpx import (
        Request as HttpxRequest,
    )
    from httpx import (
        Response as HttpxResponse,
    )
except ImportError:
    ASGITransport = None
    HttpxRequest = None
    HttpxResponse = None
    AsyncByteStream = None
    AsyncClient = None


try:
    from anyio import EndOfStream, create_memory_object_stream, create_task_group
    from mcp.shared.message import SessionMessage
    from mcp.types import (
        JSONRPCMessage,
        JSONRPCNotification,
        JSONRPCRequest,
    )

    _MCP_VERSION = package_version("mcp")
    _IS_MCP_V2 = _MCP_VERSION is not None and _MCP_VERSION >= (2, 0, 0)
except ImportError:
    create_memory_object_stream = None
    create_task_group = None
    EndOfStream = None

    JSONRPCMessage = None
    JSONRPCNotification = None
    JSONRPCRequest = None
    SessionMessage = None
    _IS_MCP_V2 = False


SENTRY_EVENT_SCHEMA = "./checkouts/data-schemas/relay/event.schema.json"

if not os.path.isfile(SENTRY_EVENT_SCHEMA):
    SENTRY_EVENT_SCHEMA = None
else:
    with open(SENTRY_EVENT_SCHEMA) as f:
        SENTRY_EVENT_SCHEMA = json.load(f)


from sentry_sdk import scope


@pytest.fixture(autouse=True)
def clean_scopes():
    """
    Resets the scopes for every test to avoid leaking data between tests.
    """
    scope._global_scope = None
    scope._isolation_scope.set(None)
    scope._current_scope.set(None)


@pytest.fixture(autouse=True)
def internal_exceptions(request):
    errors = []
    if "tests_internal_exceptions" in request.keywords:
        return

    def _capture_internal_exception(exc_info):
        errors.append(exc_info)

    @request.addfinalizer
    def _():
        # reraise the errors so that this just acts as a pass-through (that
        # happens to keep track of the errors which pass through it)
        for e in errors:
            reraise(*e)

    sentry_sdk.utils.capture_internal_exception = _capture_internal_exception

    return errors


@pytest.fixture(autouse=True, scope="session")
def _capture_internal_warnings():
    yield

    _warning_recorder_mgr.__exit__(None, None, None)
    recorder = _warning_recorder

    for warning in recorder:
        try:
            if isinstance(warning.message, ResourceWarning):
                continue
        except NameError:
            pass

        if "sentry_sdk" not in str(warning.filename) and "sentry-sdk" not in str(
            warning.filename
        ):
            continue

        # pytest-django
        if "getfuncargvalue" in str(warning.message):
            continue

        # Happens when re-initializing the SDK
        if "but it was only enabled on init()" in str(warning.message):
            continue

        # sanic's usage of aiohttp for test client
        if "verify_ssl is deprecated, use ssl=False instead" in str(warning.message):
            continue

        if "getargspec" in str(warning.message) and warning.filename.endswith(
            ("pyramid/config/util.py", "pyramid/config/views.py")
        ):
            continue

        if "isAlive() is deprecated" in str(
            warning.message
        ) and warning.filename.endswith("celery/utils/timer2.py"):
            continue

        if "collections.abc" in str(warning.message) and warning.filename.endswith(
            ("celery/canvas.py", "werkzeug/datastructures.py", "tornado/httputil.py")
        ):
            continue

        # Django 1.7 emits a (seemingly) false-positive warning for our test
        # app and suggests to use a middleware that does not exist in later
        # Django versions.
        if "SessionAuthenticationMiddleware" in str(warning.message):
            continue

        if "Something has already installed a non-asyncio" in str(warning.message):
            continue

        if "dns.hash" in str(warning.message) or "dns/namedict" in warning.filename:
            continue

        # On Python < 3.8 we fall back to importing `pkg_resources` to
        # enumerate installed modules, which emits a DeprecationWarning.
        if "pkg_resources is deprecated as an API" in str(
            warning.message
        ) and warning.filename.endswith("sentry_sdk/utils.py"):
            continue

        raise AssertionError(warning)


@pytest.fixture
def validate_event_schema(tmpdir):
    def inner(event):
        assert jsonschema is not None
        if SENTRY_EVENT_SCHEMA:
            jsonschema.validate(instance=event, schema=SENTRY_EVENT_SCHEMA)

    return inner


@pytest.fixture
def reset_integrations():
    """
    Use with caution, sometimes we really need to start
    with a clean slate to ensure monkeypatching works well,
    but this also means some other stuff will be monkeypatched twice.
    """
    global _DEFAULT_INTEGRATIONS, _processed_integrations
    try:
        _DEFAULT_INTEGRATIONS.remove(
            "sentry_sdk.integrations.opentelemetry.integration.OpenTelemetryIntegration"
        )
    except ValueError:
        pass
    _processed_integrations.clear()
    _installed_integrations.clear()


@pytest.fixture
def uninstall_integration():
    """Use to force the next call to sentry_init to re-install/setup an integration."""

    def inner(identifier):
        _processed_integrations.discard(identifier)
        _installed_integrations.discard(identifier)

    return inner


@pytest.fixture
def sentry_init(request):
    def inner(*a, **kw):
        kw.setdefault("transport", TestTransport())
        client = sentry_sdk.Client(*a, **kw)
        sentry_sdk.get_global_scope().set_client(client)

    if request.node.get_closest_marker("forked"):
        # Do not run isolation if the test is already running in
        # ultimate isolation (seems to be required for celery tests that
        # fork)
        yield inner
    else:
        old_client = sentry_sdk.get_global_scope().client
        try:
            sentry_sdk.get_current_scope().set_client(None)
            yield inner
        finally:
            sentry_sdk.get_global_scope().set_client(old_client)


class TestTransport(Transport):
    def __init__(self):
        Transport.__init__(self)

    def capture_envelope(self, _: Envelope) -> None:
        """No-op capture_envelope for tests"""
        pass


class TestTransportWithOptions(Transport):
    """TestTransport above does not pass in the options and for some tests we need them"""

    __test__ = False

    def __init__(self, options=None):
        Transport.__init__(self, options)

    def capture_envelope(self, _: Envelope) -> None:
        """No-op capture_envelope for tests"""
        pass


@pytest.fixture
def capture_events(monkeypatch):
    def inner():
        events = []
        test_client = sentry_sdk.get_client()
        old_capture_envelope = test_client.transport.capture_envelope

        def append_event(envelope):
            for item in envelope:
                if item.headers.get("type") in ("event", "transaction"):
                    events.append(item.payload.json)
            return old_capture_envelope(envelope)

        monkeypatch.setattr(test_client.transport, "capture_envelope", append_event)

        return events

    return inner


@pytest.fixture
def capture_envelopes(monkeypatch):
    def inner():
        envelopes = []
        test_client = sentry_sdk.get_client()
        old_capture_envelope = test_client.transport.capture_envelope

        def append_envelope(envelope):
            envelopes.append(envelope)
            return old_capture_envelope(envelope)

        monkeypatch.setattr(test_client.transport, "capture_envelope", append_envelope)

        return envelopes

    return inner


@dataclass
class UnwrappedItem:
    type: str
    payload: dict


@pytest.fixture
def capture_items(monkeypatch):
    """
    Capture envelope payload, unfurling individual items.

    Makes it easier to work with both events and attribute-based telemetry in
    one test.
    """

    def inner(*types):
        telemetry = []
        test_client = sentry_sdk.get_client()
        old_capture_envelope = test_client.transport.capture_envelope

        def append_envelope(envelope):
            for item in envelope:
                if types and item.type not in types:
                    continue

                if item.type in ("trace_metric", "log", "span"):
                    for i in item.payload.json["items"]:
                        t = {k: v for k, v in i.items() if k != "attributes"}
                        t["attributes"] = {
                            k: v["value"] for k, v in i["attributes"].items()
                        }
                        telemetry.append(UnwrappedItem(type=item.type, payload=t))
                else:
                    telemetry.append(
                        UnwrappedItem(type=item.type, payload=item.payload.json)
                    )

            return old_capture_envelope(envelope)

        monkeypatch.setattr(test_client.transport, "capture_envelope", append_envelope)

        return telemetry

    return inner


@pytest.fixture
def capture_record_lost_event_calls(monkeypatch):
    def inner():
        calls = []
        test_client = sentry_sdk.get_client()

        def record_lost_event(reason, data_category=None, item=None, *, quantity=1):
            calls.append((reason, data_category, item, quantity))

        monkeypatch.setattr(
            test_client.transport, "record_lost_event", record_lost_event
        )
        return calls

    return inner


@pytest.fixture
def capture_events_forksafe(monkeypatch, capture_events, request):
    def inner():
        capture_events()

        events_r, events_w = os.pipe()
        events_r = os.fdopen(events_r, "rb", 0)
        events_w = os.fdopen(events_w, "wb", 0)

        test_client = sentry_sdk.get_client()

        old_capture_envelope = test_client.transport.capture_envelope

        def append(envelope):
            event = envelope.get_event() or envelope.get_transaction_event()
            if event is not None:
                events_w.write(json.dumps(event).encode("utf-8"))
                events_w.write(b"\n")
            return old_capture_envelope(envelope)

        def flush(timeout=None, callback=None):
            events_w.write(b"flush\n")

        monkeypatch.setattr(test_client.transport, "capture_envelope", append)
        monkeypatch.setattr(test_client, "flush", flush)

        return EventStreamReader(events_r, events_w)

    return inner


@pytest.fixture
def capture_items_forksafe(monkeypatch, capture_items, request):
    def inner(*types):
        capture_items(*types)

        items_r, items_w = os.pipe()
        items_r = os.fdopen(items_r, "rb", 0)
        items_w = os.fdopen(items_w, "wb", 0)

        test_client = sentry_sdk.get_client()
        old_capture_envelope = test_client.transport.capture_envelope

        telemetry = []

        def append(envelope):
            for item in envelope:
                if types and item.type not in types:
                    continue

                if item.type in ("metric", "log", "span"):
                    for i in item.payload.json["items"]:
                        t = {k: v for k, v in i.items() if k != "attributes"}
                        t["attributes"] = {
                            k: v["value"] for k, v in i["attributes"].items()
                        }
                        telemetry.append({"type": item.type, "payload": t})
                else:
                    telemetry.append({"type": item.type, "payload": item.payload.json})

            return old_capture_envelope(envelope)

        real_flush = test_client.flush

        def flush(timeout=None, callback=None):
            real_flush(timeout=timeout, callback=callback)
            items_w.write(json.dumps(telemetry).encode("utf-8") + b"\n")
            items_w.write(b"flush\n")

        monkeypatch.setattr(test_client.transport, "capture_envelope", append)
        monkeypatch.setattr(test_client, "flush", flush)

        return EventStreamReader(items_r, items_w)

    return inner


class EventStreamReader:
    def __init__(self, read_file, write_file):
        self.read_file = read_file
        self.write_file = write_file

    def read_event(self):
        return json.loads(self.read_file.readline().decode("utf-8"))

    def read_flush(self):
        assert self.read_file.readline() == b"flush\n"


# scope=session ensures that fixture is run earlier
@pytest.fixture(
    scope="session",
    params=[None, "eventlet", "gevent"],
    ids=("threads", "eventlet", "greenlet"),
)
def maybe_monkeypatched_threading(request):
    if request.param == "eventlet":
        if eventlet is None:
            pytest.skip("no eventlet installed")

        try:
            eventlet.monkey_patch()
        except AttributeError as e:
            if "'thread.RLock' object has no attribute" in str(e):
                # https://bitbucket.org/pypy/pypy/issues/2962/gevent-cannot-patch-rlock-under-pypy-27-7
                pytest.skip("https://github.com/eventlet/eventlet/issues/546")
            else:
                raise
    elif request.param == "gevent":
        if gevent is None:
            pytest.skip("no gevent installed")
        try:
            gevent.monkey.patch_all()
        except Exception as e:
            if "_RLock__owner" in str(e):
                pytest.skip("https://github.com/gevent/gevent/issues/1380")
            else:
                raise
    else:
        assert request.param is None

    return request.param


@pytest.fixture
def render_span_tree():
    def inner(spans, root_span=None):
        streamed_spans = False
        if root_span is None:
            streamed_spans = True

        by_parent = {}
        for span in spans:
            if "parent_span_id" not in span:
                root_span = span
                continue

            by_parent.setdefault(span["parent_span_id"], []).append(span)

        def render_span(span):
            if streamed_spans:
                yield "- sentry.op={}: name={}".format(
                    json.dumps(span["attributes"].get("sentry.op")),
                    json.dumps(span["name"]),
                )
            else:
                yield "- op={}: description={}".format(
                    json.dumps(span.get("op")), json.dumps(span.get("description"))
                )

            for subspan in by_parent.get(span["span_id"]) or ():
                for line in render_span(subspan):
                    yield "  {}".format(line)

        return "\n".join(render_span(root_span))

    return inner


@pytest.fixture(name="StringContaining")
def string_containing_matcher():
    """
    An object which matches any string containing the substring passed to the
    object at instantiation time.

    Useful for assert_called_with, assert_any_call, etc.

    Used like this:

    >>> f = mock.Mock()
    >>> f("dogs are great")
    >>> f.assert_any_call("dogs") # will raise AssertionError
    Traceback (most recent call last):
        ...
    AssertionError: mock('dogs') call not found
    >>> f.assert_any_call(StringContaining("dogs")) # no AssertionError

    """

    class StringContaining:
        def __init__(self, substring):
            self.substring = substring
            self.valid_types = (str, bytes)

        def __eq__(self, test_string):
            if not isinstance(test_string, self.valid_types):
                return False

            # this is safe even in py2 because as of 2.6, `bytes` exists in py2
            # as an alias for `str`
            if isinstance(test_string, bytes):
                test_string = test_string.decode()

            if len(self.substring) > len(test_string):
                return False

            return self.substring in test_string

        def __ne__(self, test_string):
            return not self.__eq__(test_string)

    return StringContaining


def _safe_is_equal(x, y):
    """
    Compares two values, preferring to use the first's __eq__ method if it
    exists and is implemented.

    Accounts for py2/py3 differences (like ints in py2 not having a __eq__
    method), as well as the incomparability of certain types exposed by using
    raw __eq__ () rather than ==.
    """

    # Prefer using __eq__ directly to ensure that examples like
    #
    #   maisey = Dog()
    #   maisey.name = "Maisey the Dog"
    #   maisey == ObjectDescribedBy(attrs={"name": StringContaining("Maisey")})
    #
    # evaluate to True (in other words, examples where the values in self.attrs
    # might also have custom __eq__ methods; this makes sure those methods get
    # used if possible)
    try:
        is_equal = x.__eq__(y)
    except AttributeError:
        is_equal = NotImplemented

    # this can happen on its own, too (i.e. without an AttributeError being
    # thrown), which is why this is separate from the except block above
    if is_equal == NotImplemented:
        # using == smoothes out weird variations exposed by raw __eq__
        return x == y

    return is_equal


@pytest.fixture(name="DictionaryContaining")
def dictionary_containing_matcher():
    """
    An object which matches any dictionary containing all key-value pairs from
    the dictionary passed to the object at instantiation time.

    Useful for assert_called_with, assert_any_call, etc.

    Used like this:

    >>> f = mock.Mock()
    >>> f({"dogs": "yes", "cats": "maybe"})
    >>> f.assert_any_call({"dogs": "yes"}) # will raise AssertionError
    Traceback (most recent call last):
        ...
    AssertionError: mock({'dogs': 'yes'}) call not found
    >>> f.assert_any_call(DictionaryContaining({"dogs": "yes"})) # no AssertionError
    """

    class DictionaryContaining:
        def __init__(self, subdict):
            self.subdict = subdict

        def __eq__(self, test_dict):
            if not isinstance(test_dict, dict):
                return False

            if len(self.subdict) > len(test_dict):
                return False

            for key, value in self.subdict.items():
                try:
                    test_value = test_dict[key]
                except KeyError:  # missing key
                    return False

                if not _safe_is_equal(value, test_value):
                    return False

            return True

        def __ne__(self, test_dict):
            return not self.__eq__(test_dict)

    return DictionaryContaining


@pytest.fixture(name="ObjectDescribedBy")
def object_described_by_matcher():
    """
    An object which matches any other object with the given properties.

    Available properties currently are "type" (a type object) and "attrs" (a
    dictionary).

    Useful for assert_called_with, assert_any_call, etc.

    Used like this:

    >>> class Dog:
    ...     pass
    ...
    >>> maisey = Dog()
    >>> maisey.name = "Maisey"
    >>> maisey.age = 7
    >>> f = mock.Mock()
    >>> f(maisey)
    >>> f.assert_any_call(ObjectDescribedBy(type=Dog)) # no AssertionError
    >>> f.assert_any_call(ObjectDescribedBy(attrs={"name": "Maisey"})) # no AssertionError
    """

    class ObjectDescribedBy:
        def __init__(self, type=None, attrs=None):
            self.type = type
            self.attrs = attrs

        def __eq__(self, test_obj):
            if self.type:
                if not isinstance(test_obj, self.type):
                    return False

            if self.attrs:
                for attr_name, attr_value in self.attrs.items():
                    try:
                        test_value = getattr(test_obj, attr_name)
                    except AttributeError:  # missing attribute
                        return False

                    if not _safe_is_equal(attr_value, test_value):
                        return False

            return True

        def __ne__(self, test_obj):
            return not self.__eq__(test_obj)

    return ObjectDescribedBy


@pytest.fixture
def teardown_profiling():
    # Make sure that a previous test didn't leave the profiler running
    teardown_profiler()
    teardown_continuous_profiler()

    yield

    # Make sure that to shut down the profiler after the test
    teardown_profiler()
    teardown_continuous_profiler()


@pytest.fixture()
def suppress_deprecation_warnings():
    """
    Use this fixture to suppress deprecation warnings in a test.
    Useful for testing deprecated SDK features.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        yield


def _make_session_message(jsonrpc_msg):
    """Construct a SessionMessage compatible with both MCP v1 and v2."""
    if _IS_MCP_V2:
        return SessionMessage(message=jsonrpc_msg)  # type: ignore
    else:
        return SessionMessage(message=JSONRPCMessage(root=jsonrpc_msg))  # type: ignore


@pytest.fixture
def get_initialization_payload():
    def inner(request_id: str):
        return _make_session_message(
            JSONRPCRequest(  # type: ignore
                jsonrpc="2.0",
                id=request_id,
                method="initialize",
                params={
                    "protocolVersion": "2025-11-25",
                    "capabilities": {},
                    "clientInfo": {"name": "test-client", "version": "1.0.0"},
                },
            )
        )

    return inner


@pytest.fixture
def get_initialized_notification_payload():
    def inner():
        return _make_session_message(
            JSONRPCNotification(  # type: ignore
                jsonrpc="2.0",
                method="notifications/initialized",
            )
        )

    return inner


@pytest.fixture
def get_mcp_command_payload():
    def inner(method: str, params, request_id: str):
        return _make_session_message(
            JSONRPCRequest(  # type: ignore
                jsonrpc="2.0",
                id=request_id,
                method=method,
                params=params,
            )
        )

    return inner


@pytest.fixture
def stdio(
    get_initialization_payload,
    get_initialized_notification_payload,
    get_mcp_command_payload,
):
    async def inner(server, method: str, params, request_id: str | None = None):
        if request_id is None:
            request_id = "1"

        read_stream_writer, read_stream = create_memory_object_stream(0)  # type: ignore
        write_stream, write_stream_reader = create_memory_object_stream(0)  # type: ignore

        result = {}

        async def run_server():
            await server.run(
                read_stream, write_stream, server.create_initialization_options()
            )

        async def simulate_client(tg, result):
            init_request = get_initialization_payload("1")
            await read_stream_writer.send(init_request)

            await write_stream_reader.receive()

            initialized_notification = get_initialized_notification_payload()
            await read_stream_writer.send(initialized_notification)

            request = get_mcp_command_payload(
                method, params=params, request_id=request_id
            )
            await read_stream_writer.send(request)

            result["response"] = await write_stream_reader.receive()

            tg.cancel_scope.cancel()

        async with create_task_group() as tg:  # type: ignore
            tg.start_soon(run_server)
            tg.start_soon(simulate_client, tg, result)

        return result["response"]

    return inner


@pytest.fixture()
def json_rpc():
    def inner(app, method: str, params, request_id: str):
        with TestClient(app) as client:  # type: ignore
            init_response = client.post(
                "/mcp/",
                headers={
                    "Accept": "application/json, text/event-stream",
                    "Content-Type": "application/json",
                },
                json={
                    "jsonrpc": "2.0",
                    "method": "initialize",
                    "params": {
                        "clientInfo": {"name": "test-client", "version": "1.0"},
                        "protocolVersion": "2025-11-25",
                        "capabilities": {},
                    },
                    "id": request_id,
                },
            )

            session_id = init_response.headers["mcp-session-id"]

            # Notification response is mandatory.
            # https://modelcontextprotocol.io/specification/2025-11-25/basic/lifecycle
            client.post(
                "/mcp/",
                headers={
                    "Accept": "application/json, text/event-stream",
                    "Content-Type": "application/json",
                    "mcp-session-id": session_id,
                },
                json={
                    "jsonrpc": "2.0",
                    "method": "notifications/initialized",
                    "params": {},
                },
            )

            response = client.post(
                "/mcp/",
                headers={
                    "Accept": "application/json, text/event-stream",
                    "Content-Type": "application/json",
                    "mcp-session-id": session_id,
                },
                json={
                    "jsonrpc": "2.0",
                    "method": method,
                    "params": params,
                    "id": request_id,
                },
            )

            return session_id, response

    return inner


@pytest.fixture()
def select_mcp_transactions():
    def inner(events):
        return [
            event
            for event in events
            if event["type"] == "transaction"
            and event["contexts"]["trace"]["op"] == "mcp.server"
        ]

    return inner


@pytest.fixture()
def select_transactions_with_mcp_spans():
    def inner(events, method_name):
        return [
            transaction
            for transaction in events
            if transaction["type"] == "transaction"
            and any(
                span["data"].get("mcp.method.name") == method_name
                for span in transaction.get("spans", [])
            )
        ]

    return inner


@pytest.fixture()
def json_rpc_sse():
    class StreamingASGITransport(ASGITransport):
        """
        Simple transport whose only purpose is to keep GET request alive in SSE connections, allowing
        tests involving SSE interactions to run in-process.
        """

        def __init__(
            self,
            app: "Callable",
            keep_sse_alive: "asyncio.Event",
        ) -> None:
            self.keep_sse_alive = keep_sse_alive
            super().__init__(app)

        async def handle_async_request(
            self, request: "HttpxRequest"
        ) -> "HttpxResponse":
            scope = {
                "type": "http",
                "method": request.method,
                "headers": [(k.lower(), v) for (k, v) in request.headers.raw],
                "path": request.url.path,
                "query_string": request.url.query,
            }

            is_streaming_sse = scope["method"] == "GET" and scope["path"] == "/sse"
            if not is_streaming_sse:
                return await super().handle_async_request(request)

            request_body = b""
            if request.content:
                request_body = await request.aread()

            body_sender, body_receiver = create_memory_object_stream[bytes](0)  # type: ignore

            async def receive() -> "dict[str, Any]":
                if self.keep_sse_alive.is_set():
                    return {"type": "http.disconnect"}

                await self.keep_sse_alive.wait()  # Keep alive :)
                return {
                    "type": "http.request",
                    "body": request_body,
                    "more_body": False,
                }

            async def send(message: "MutableMapping[str, Any]") -> None:
                if message["type"] == "http.response.body":
                    body = message.get("body", b"")
                    more_body = message.get("more_body", False)

                    if body == b"" and not more_body:
                        return

                    if body:
                        await body_sender.send(body)

                    if not more_body:
                        await body_sender.aclose()

            async def run_app():
                await self.app(scope, receive, send)

            class StreamingBodyStream(AsyncByteStream):  # type: ignore
                def __init__(self, receiver):
                    self.receiver = receiver

                async def __aiter__(self):
                    try:
                        async for chunk in self.receiver:
                            yield chunk
                    except EndOfStream:  # type: ignore
                        pass

            stream = StreamingBodyStream(body_receiver)
            response = HttpxResponse(status_code=200, headers=[], stream=stream)  # type: ignore

            asyncio.create_task(run_app())
            return response

    def parse_sse_data_package(sse_chunk):
        sse_text = sse_chunk.decode("utf-8")
        json_str = sse_text.split("data: ")[1]
        return json.loads(json_str)

    async def inner(
        app, method: str, params, request_id: str, keep_sse_alive: "asyncio.Event"
    ):
        context = {}

        stream_complete = asyncio.Event()
        endpoint_parsed = asyncio.Event()

        # https://github.com/Kludex/starlette/issues/104#issuecomment-729087925
        async with AsyncClient(  # type: ignore
            transport=StreamingASGITransport(app=app, keep_sse_alive=keep_sse_alive),
            base_url="http://test",
        ) as client:

            async def parse_stream():
                async with client.stream("GET", "/sse") as stream:
                    # Read directly from stream.stream instead of aiter_bytes()
                    async for chunk in stream.stream:
                        if b"event: endpoint" in chunk:
                            sse_text = chunk.decode("utf-8")
                            url = sse_text.split("data: ")[1]

                            parsed = urlparse(url)
                            query_params = parse_qs(parsed.query)
                            context["session_id"] = query_params["session_id"][0]
                            endpoint_parsed.set()
                            continue

                        if b"event: message" in chunk and b"structuredContent" in chunk:
                            context["response"] = parse_sse_data_package(chunk)
                            break
                        elif (
                            "result" in parse_sse_data_package(chunk)
                            and "content" in parse_sse_data_package(chunk)["result"]
                        ):
                            context["response"] = parse_sse_data_package(chunk)
                            break

                stream_complete.set()

            task = asyncio.create_task(parse_stream())
            await endpoint_parsed.wait()

            await client.post(
                f"/messages/?session_id={context['session_id']}",
                headers={
                    "Content-Type": "application/json",
                },
                json={
                    "jsonrpc": "2.0",
                    "method": "initialize",
                    "params": {
                        "clientInfo": {"name": "test-client", "version": "1.0"},
                        "protocolVersion": "2025-11-25",
                        "capabilities": {},
                    },
                    "id": request_id,
                },
            )

            # Notification response is mandatory.
            # https://modelcontextprotocol.io/specification/2025-11-25/basic/lifecycle
            await client.post(
                f"/messages/?session_id={context['session_id']}",
                headers={
                    "Content-Type": "application/json",
                    "mcp-session-id": context["session_id"],
                },
                json={
                    "jsonrpc": "2.0",
                    "method": "notifications/initialized",
                    "params": {},
                },
            )

            await client.post(
                f"/messages/?session_id={context['session_id']}",
                headers={
                    "Content-Type": "application/json",
                    "mcp-session-id": context["session_id"],
                },
                json={
                    "jsonrpc": "2.0",
                    "method": method,
                    "params": params,
                    "id": request_id,
                },
            )

            await stream_complete.wait()
            keep_sse_alive.set()

            return task, context["session_id"], context["response"]

    return inner


@pytest.fixture()
def async_iterator():
    async def inner(values):
        for value in values:
            yield value

    return inner


@pytest.fixture
def server_side_event_chunks():
    def inner(events, include_event_type=True):
        for event in events:
            payload = event.model_dump()
            chunk = (
                f"event: {payload['type']}\ndata: {json.dumps(payload)}\n\n"
                if include_event_type
                else f"data: {json.dumps(payload)}\n\n"
            )
            yield chunk.encode("utf-8")

    return inner


@pytest.fixture
def get_model_response():
    def inner(response_content, serialize_pydantic=False, request_headers=None):
        if request_headers is None:
            request_headers = {}

        model_request = HttpxRequest(
            "POST",
            "/responses",
            headers=request_headers,
        )

        if serialize_pydantic:
            response_content = json.dumps(
                response_content.model_dump(
                    by_alias=True,
                    exclude_none=True,
                )
            ).encode("utf-8")

        response = HttpxResponse(
            200,
            request=model_request,
            content=response_content,
        )

        return response

    return inner


@pytest.fixture
def get_rate_limit_model_response():
    def inner(request_headers=None):
        if request_headers is None:
            request_headers = {}

        model_request = HttpxRequest(
            "POST",
            "/responses",
            headers=request_headers,
        )

        response = HttpxResponse(
            429,
            request=model_request,
        )

        return response

    return inner


@pytest.fixture
def streaming_chat_completions_model_response():
    return [
        openai.types.chat.ChatCompletionChunk(
            id="chatcmpl-test",
            object="chat.completion.chunk",
            created=10000000,
            model="gpt-3.5-turbo",
            choices=[
                openai.types.chat.chat_completion_chunk.Choice(
                    index=0,
                    delta=openai.types.chat.chat_completion_chunk.ChoiceDelta(
                        role="assistant"
                    ),
                    finish_reason=None,
                ),
            ],
        ),
        openai.types.chat.ChatCompletionChunk(
            id="chatcmpl-test",
            object="chat.completion.chunk",
            created=10000000,
            model="gpt-3.5-turbo",
            choices=[
                openai.types.chat.chat_completion_chunk.Choice(
                    index=0,
                    delta=openai.types.chat.chat_completion_chunk.ChoiceDelta(
                        content="Tes"
                    ),
                    finish_reason=None,
                ),
            ],
        ),
        openai.types.chat.ChatCompletionChunk(
            id="chatcmpl-test",
            object="chat.completion.chunk",
            created=10000000,
            model="gpt-3.5-turbo",
            choices=[
                openai.types.chat.chat_completion_chunk.Choice(
                    index=0,
                    delta=openai.types.chat.chat_completion_chunk.ChoiceDelta(
                        content="t r"
                    ),
                    finish_reason=None,
                ),
            ],
        ),
        openai.types.chat.ChatCompletionChunk(
            id="chatcmpl-test",
            object="chat.completion.chunk",
            created=10000000,
            model="gpt-3.5-turbo",
            choices=[
                openai.types.chat.chat_completion_chunk.Choice(
                    index=0,
                    delta=openai.types.chat.chat_completion_chunk.ChoiceDelta(
                        content="esp"
                    ),
                    finish_reason=None,
                ),
            ],
        ),
        openai.types.chat.ChatCompletionChunk(
            id="chatcmpl-test",
            object="chat.completion.chunk",
            created=10000000,
            model="gpt-3.5-turbo",
            choices=[
                openai.types.chat.chat_completion_chunk.Choice(
                    index=0,
                    delta=openai.types.chat.chat_completion_chunk.ChoiceDelta(
                        content="ons"
                    ),
                    finish_reason=None,
                ),
            ],
        ),
        openai.types.chat.ChatCompletionChunk(
            id="chatcmpl-test",
            object="chat.completion.chunk",
            created=10000000,
            model="gpt-3.5-turbo",
            choices=[
                openai.types.chat.chat_completion_chunk.Choice(
                    index=0,
                    delta=openai.types.chat.chat_completion_chunk.ChoiceDelta(
                        content="e"
                    ),
                    finish_reason=None,
                ),
            ],
        ),
        openai.types.chat.ChatCompletionChunk(
            id="chatcmpl-test",
            object="chat.completion.chunk",
            created=10000000,
            model="gpt-3.5-turbo",
            choices=[
                openai.types.chat.chat_completion_chunk.Choice(
                    index=0,
                    delta=openai.types.chat.chat_completion_chunk.ChoiceDelta(),
                    finish_reason="stop",
                ),
            ],
            usage=openai.types.CompletionUsage(
                prompt_tokens=10,
                completion_tokens=20,
                total_tokens=30,
            ),
        ),
    ]


@pytest.fixture
def nonstreaming_chat_completions_model_response():
    def inner(
        response_id: str,
        response_model: str,
        message_content: str,
        created: int,
        usage: openai.types.CompletionUsage,
    ):
        return openai.types.chat.ChatCompletion(
            id=response_id,
            choices=[
                openai.types.chat.chat_completion.Choice(
                    index=0,
                    finish_reason="stop",
                    message=openai.types.chat.ChatCompletionMessage(
                        role="assistant", content=message_content
                    ),
                )
            ],
            created=created,
            model=response_model,
            object="chat.completion",
            usage=usage,
        )

    return inner


@pytest.fixture
def openai_embedding_model_response():
    return openai.types.CreateEmbeddingResponse(
        data=[
            openai.types.Embedding(
                embedding=[0.1, 0.2, 0.3],
                index=0,
                object="embedding",
            )
        ],
        model="text-embedding-ada-002",
        object="list",
        usage=openai.types.create_embedding_response.Usage(
            prompt_tokens=5,
            total_tokens=5,
        ),
    )


@pytest.fixture
def nonstreaming_responses_model_response():
    return openai.types.responses.Response(
        id="resp_123",
        output=[
            openai.types.responses.ResponseOutputMessage(
                id="msg_123",
                type="message",
                status="completed",
                content=[
                    openai.types.responses.ResponseOutputText(
                        text="Hello, how can I help you?",
                        type="output_text",
                        annotations=[],
                    )
                ],
                role="assistant",
            )
        ],
        parallel_tool_calls=False,
        tool_choice="none",
        tools=[],
        created_at=10000000,
        model="gpt-4",
        object="response",
        usage=openai.types.responses.ResponseUsage(
            input_tokens=10,
            input_tokens_details=openai.types.responses.response_usage.InputTokensDetails(
                cached_tokens=0,
                cache_write_tokens=0,
            ),
            output_tokens=20,
            output_tokens_details=openai.types.responses.response_usage.OutputTokensDetails(
                reasoning_tokens=5,
            ),
            total_tokens=30,
        ),
    )


@pytest.fixture
def nonstreaming_anthropic_model_response():
    return anthropic.types.Message(
        id="msg_123",
        type="message",
        role="assistant",
        model="claude-3-opus-20240229",
        content=[
            anthropic.types.TextBlock(
                type="text",
                text="Hello, how can I help you?",
            )
        ],
        stop_reason="end_turn",
        stop_sequence=None,
        usage=anthropic.types.Usage(
            input_tokens=10,
            output_tokens=20,
        ),
    )


@pytest.fixture
def nonstreaming_google_genai_model_response():
    return google.genai.types.GenerateContentResponse(
        response_id="resp_123",
        candidates=[
            google.genai.types.Candidate(
                content=google.genai.types.Content(
                    role="model",
                    parts=[
                        google.genai.types.Part(
                            text="Hello, how can I help you?",
                        )
                    ],
                ),
                finish_reason="STOP",
            )
        ],
        model_version="gemini/gemini-pro",
        usage_metadata=google.genai.types.GenerateContentResponseUsageMetadata(
            prompt_token_count=10,
            candidates_token_count=20,
            total_token_count=30,
        ),
    )


@pytest.fixture
def responses_tool_call_model_responses():
    def inner(
        tool_name: str,
        arguments: str,
        response_model: str,
        response_text: str,
        response_ids: "Iterator[str]",
        usages: "Iterator[openai.types.responses.ResponseUsage]",
    ):
        yield openai.types.responses.Response(
            id=next(response_ids),
            output=[
                openai.types.responses.ResponseFunctionToolCall(
                    id="call_123",
                    call_id="call_123",
                    name=tool_name,
                    type="function_call",
                    arguments=arguments,
                )
            ],
            parallel_tool_calls=False,
            tool_choice="none",
            tools=[],
            created_at=10000000,
            model=response_model,
            object="response",
            usage=next(usages),
        )

        yield openai.types.responses.Response(
            id=next(response_ids),
            output=[
                openai.types.responses.ResponseOutputMessage(
                    id="msg_final",
                    type="message",
                    status="completed",
                    content=[
                        openai.types.responses.ResponseOutputText(
                            text=response_text,
                            type="output_text",
                            annotations=[],
                        )
                    ],
                    role="assistant",
                )
            ],
            parallel_tool_calls=False,
            tool_choice="none",
            tools=[],
            created_at=10000000,
            model=response_model,
            object="response",
            usage=next(usages),
        )

    return inner


class MockServerRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):  # noqa: N802
        # Process an HTTP GET request and return a response.
        # If the path ends with /status/<number>, return status code <number>.
        # Otherwise return a 200 response.
        code = 200
        if "/status/" in self.path:
            code = int(self.path[-3:])

        self.send_response(code)
        self.end_headers()
        return


def get_free_port():
    s = socket.socket(socket.AF_INET, type=socket.SOCK_STREAM)
    s.bind(("localhost", 0))
    _, port = s.getsockname()
    s.close()
    return port


def create_mock_http_server():
    # Start a mock server to test outgoing http requests
    mock_server_port = get_free_port()
    mock_server = HTTPServer(("localhost", mock_server_port), MockServerRequestHandler)
    mock_server_thread = Thread(target=mock_server.serve_forever)
    mock_server_thread.daemon = True
    mock_server_thread.start()

    return mock_server_port


def unpack_werkzeug_response(response):
    # werkzeug < 2.1 returns a tuple as client response, newer versions return
    # an object
    try:
        return response.get_data(), response.status, response.headers
    except AttributeError:
        content, status, headers = response
        return b"".join(content), status, headers


def werkzeug_set_cookie(client, servername, key, value):
    # client.set_cookie has a different signature in different werkzeug versions
    try:
        client.set_cookie(servername, key, value)
    except TypeError:
        client.set_cookie(key, value)


@contextmanager
def patch_start_tracing_child(
    fake_transaction_is_none: bool = False,
) -> "Iterator[Optional[mock.MagicMock]]":
    if not fake_transaction_is_none:
        fake_transaction = mock.MagicMock()
        fake_start_child = mock.MagicMock()
        fake_transaction.start_child = fake_start_child
    else:
        fake_transaction = None
        fake_start_child = None

    with mock.patch(
        "sentry_sdk.tracing_utils.get_current_span", return_value=fake_transaction
    ):
        yield fake_start_child


class ApproxDict(dict):
    def __eq__(self, other):
        # For an ApproxDict to equal another dict, the other dict just needs to contain
        # all the keys from the ApproxDict with the same values.
        #
        # The other dict may contain additional keys with any value.
        return all(key in other and other[key] == value for key, value in self.items())

    def __ne__(self, other):
        return not self.__eq__(other)


CapturedData = namedtuple("CapturedData", ["path", "event", "envelope", "compressed"])


@pytest.fixture(scope="module")
def wsgi_capturing_server():
    assert WSGIServer is not None

    class CapturingServer(WSGIServer):
        def __init__(self, host="127.0.0.1", port=0, ssl_context=None):
            WSGIServer.__init__(self, host, port, self, ssl_context=ssl_context)
            self.code = 204
            self.headers = {}
            self.captured = []

        def respond_with(self, code=200, headers=None):
            self.code = code
            if headers:
                self.headers = headers

        def clear_captured(self):
            del self.captured[:]

        def __call__(self, environ, start_response):
            """
            This is the WSGI application.
            """
            assert Request is not None
            assert Response is not None

            request = Request(environ)
            event = envelope = None
            content_encoding = request.headers.get("content-encoding")
            if content_encoding == "gzip":
                rdr = gzip.GzipFile(fileobj=io.BytesIO(request.data))
                compressed = True
            elif content_encoding == "br":
                rdr = io.BytesIO(brotli.decompress(request.data))
                compressed = True
            else:
                rdr = io.BytesIO(request.data)
                compressed = False

            if request.mimetype == "application/json":
                event = parse_json(rdr.read())
            else:
                envelope = Envelope.deserialize_from(rdr)

            self.captured.append(
                CapturedData(
                    path=request.path,
                    event=event,
                    envelope=envelope,
                    compressed=compressed,
                )
            )

            response = Response(status=self.code)
            response.headers.extend(self.headers)
            return response(environ, start_response)

    return CapturingServer()



from django.core import signals
from django.dispatch import receiver

myapp_custom_signal = signals.Signal()
myapp_custom_signal_silenced = signals.Signal()


@receiver(myapp_custom_signal)
def signal_handler(sender, **kwargs):
    assert sender == "hello"


@receiver(myapp_custom_signal_silenced)
def signal_handler_silenced(sender, **kwargs):
    assert sender == "hello"



from functools import partial, partialmethod

from sentry_sdk.utils import transaction_from_function


class MyClass:
    def myfunc(self):
        pass


def myfunc():
    pass


@partial
def my_partial():
    pass


my_lambda = lambda: None

my_partial_lambda = partial(lambda: None)


def test_transaction_from_function():
    x = transaction_from_function
    assert x(MyClass) == "tests.utils.test_transaction.MyClass"
    assert x(MyClass.myfunc) == "tests.utils.test_transaction.MyClass.myfunc"
    assert x(myfunc) == "tests.utils.test_transaction.myfunc"
    assert x(None) is None
    assert x(42) is None
    assert x(lambda: None).endswith("<lambda>")
    assert x(my_lambda) == "tests.utils.test_transaction.<lambda>"
    assert (
        x(my_partial) == "partial(<function tests.utils.test_transaction.my_partial>)"
    )
    assert (
        x(my_partial_lambda)
        == "partial(<function tests.utils.test_transaction.<lambda>>)"
    )


def test_transaction_from_function_partialmethod():
    x = transaction_from_function

    class MyPartialClass:
        @partialmethod
        def my_partial_method(self):
            pass

    assert (
        x(MyPartialClass.my_partial_method)
        == "partialmethod(<function tests.utils.test_transaction.test_transaction_from_function_partialmethod.<locals>.MyPartialClass.my_partial_method>)"
    )



import pytest
from fastapi import FastAPI, Request
from fastapi.testclient import TestClient
from flask import Flask, jsonify, request
from graphene import ObjectType, Schema, String

import sentry_sdk
from sentry_sdk.consts import OP
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.flask import FlaskIntegration
from sentry_sdk.integrations.graphene import GrapheneIntegration
from sentry_sdk.integrations.starlette import StarletteIntegration


class Query(ObjectType):
    hello = String(first_name=String(default_value="stranger"))
    goodbye = String()

    def resolve_hello(root, info, first_name):  # noqa: N805
        return "Hello {}!".format(first_name)

    def resolve_goodbye(root, info):  # noqa: N805
        raise RuntimeError("oh no!")


def test_capture_request_if_available_and_send_pii_is_on_async(
    sentry_init, capture_events
):
    sentry_init(
        send_default_pii=True,
        integrations=[
            GrapheneIntegration(),
            FastApiIntegration(),
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    schema = Schema(query=Query)

    async_app = FastAPI()

    @async_app.post("/graphql")
    async def graphql_server_async(request: Request):
        data = await request.json()
        result = await schema.execute_async(data["query"])
        return result.data

    query = {"query": "query ErrorQuery {goodbye}"}
    client = TestClient(async_app)
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "graphene"
    assert event["request"]["api_target"] == "graphql"
    assert event["request"]["data"] == query


def test_capture_request_if_available_and_send_pii_is_on_sync(
    sentry_init, capture_events
):
    sentry_init(
        send_default_pii=True,
        integrations=[GrapheneIntegration(), FlaskIntegration()],
    )
    events = capture_events()

    schema = Schema(query=Query)

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server_sync():
        data = request.get_json()
        result = schema.execute(data["query"])
        return jsonify(result.data), 200

    query = {"query": "query ErrorQuery {goodbye}"}
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "graphene"
    assert event["request"]["api_target"] == "graphql"
    assert event["request"]["data"] == query


def test_do_not_capture_request_if_send_pii_is_off_async(sentry_init, capture_events):
    sentry_init(
        integrations=[
            GrapheneIntegration(),
            FastApiIntegration(),
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    schema = Schema(query=Query)

    async_app = FastAPI()

    @async_app.post("/graphql")
    async def graphql_server_async(request: Request):
        data = await request.json()
        result = await schema.execute_async(data["query"])
        return result.data

    query = {"query": "query ErrorQuery {goodbye}"}
    client = TestClient(async_app)
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "graphene"
    assert "data" not in event["request"]
    assert "response" not in event["contexts"]


def test_do_not_capture_request_if_send_pii_is_off_sync(sentry_init, capture_events):
    sentry_init(
        integrations=[GrapheneIntegration(), FlaskIntegration()],
    )
    events = capture_events()

    schema = Schema(query=Query)

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server_sync():
        data = request.get_json()
        result = schema.execute(data["query"])
        return jsonify(result.data), 200

    query = {"query": "query ErrorQuery {goodbye}"}
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "graphene"
    assert "data" not in event["request"]
    assert "response" not in event["contexts"]


def test_no_event_if_no_errors_async(sentry_init, capture_events):
    sentry_init(
        integrations=[
            GrapheneIntegration(),
            FastApiIntegration(),
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    schema = Schema(query=Query)

    async_app = FastAPI()

    @async_app.post("/graphql")
    async def graphql_server_async(request: Request):
        data = await request.json()
        result = await schema.execute_async(data["query"])
        return result.data

    query = {
        "query": "query GreetingQuery { hello }",
    }
    client = TestClient(async_app)
    client.post("/graphql", json=query)

    assert len(events) == 0


def test_no_event_if_no_errors_sync(sentry_init, capture_events):
    sentry_init(
        integrations=[
            GrapheneIntegration(),
            FlaskIntegration(),
        ],
    )
    events = capture_events()

    schema = Schema(query=Query)

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server_sync():
        data = request.get_json()
        result = schema.execute(data["query"])
        return jsonify(result.data), 200

    query = {
        "query": "query GreetingQuery { hello }",
    }
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 0


@pytest.mark.parametrize(
    "send_default_pii",
    [True, False],
)
def test_graphql_span_holds_query_information(
    sentry_init, capture_events, send_default_pii
):
    sentry_init(
        integrations=[GrapheneIntegration(), FlaskIntegration()],
        traces_sample_rate=1.0,
        default_integrations=False,
        send_default_pii=send_default_pii,
    )
    events = capture_events()

    schema = Schema(query=Query)

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server_sync():
        data = request.get_json()
        result = schema.execute(data["query"], operation_name=data.get("operationName"))
        return jsonify(result.data), 200

    query = {
        "query": "query GreetingQuery { hello }",
        "operationName": "GreetingQuery",
    }
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert len(event["spans"]) == 1

    (span,) = event["spans"]
    assert span["op"] == OP.GRAPHQL_QUERY
    assert span["description"] == query["operationName"]
    assert span["data"]["graphql.operation.name"] == query["operationName"]
    assert span["data"]["graphql.operation.type"] == "query"

    if send_default_pii is True:
        assert span["data"]["graphql.document"] == query["query"]
    else:
        assert "graphql.document" not in span["data"]


@pytest.mark.parametrize(
    "send_default_pii",
    [True, False],
)
def test_graphql_streamed_span_holds_query_information(
    sentry_init, capture_items, send_default_pii
):
    sentry_init(
        integrations=[GrapheneIntegration(), FlaskIntegration()],
        traces_sample_rate=1.0,
        default_integrations=False,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    schema = Schema(query=Query)

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server_sync():
        data = request.get_json()
        result = schema.execute(data["query"], operation_name=data.get("operationName"))
        return jsonify(result.data), 200

    query = {
        "query": "query GreetingQuery { hello }",
        "operationName": "GreetingQuery",
    }
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    sentry_sdk.get_client().flush()

    spans = [item.payload for item in items]
    assert len(spans) == 2

    graphql_span, flask_segment = spans

    assert graphql_span["name"] == query["operationName"]
    assert graphql_span["attributes"]["sentry.op"] == OP.GRAPHQL_QUERY
    assert (
        graphql_span["attributes"]["graphql.operation.name"] == query["operationName"]
    )
    assert graphql_span["attributes"]["graphql.operation.type"] == "query"
    assert graphql_span["is_segment"] is False

    if send_default_pii is True:
        assert graphql_span["attributes"]["graphql.document"] == query["query"]
    else:
        assert "graphql.document" not in graphql_span["attributes"]

    assert flask_segment["is_segment"] is True
    assert graphql_span["parent_span_id"] == flask_segment["span_id"]


def test_breadcrumbs_hold_query_information_on_error(sentry_init, capture_events):
    sentry_init(
        integrations=[
            GrapheneIntegration(),
        ],
        default_integrations=False,
    )
    events = capture_events()

    schema = Schema(query=Query)

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server_sync():
        data = request.get_json()
        result = schema.execute(data["query"], operation_name=data.get("operationName"))
        return jsonify(result.data), 200

    query = {
        "query": "query ErrorQuery { goodbye }",
        "operationName": "ErrorQuery",
    }
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert len(event["breadcrumbs"]) == 1

    breadcrumbs = event["breadcrumbs"]["values"]
    assert len(breadcrumbs) == 1

    (breadcrumb,) = breadcrumbs
    assert breadcrumb["category"] == "graphql.operation"
    assert breadcrumb["data"]["operation_name"] == query["operationName"]
    assert breadcrumb["data"]["operation_type"] == "query"
    assert breadcrumb["type"] == "default"


def test_breadcrumbs_hold_query_information_on_error_with_span_streaming(
    sentry_init, capture_items
):
    sentry_init(
        integrations=[
            GrapheneIntegration(),
        ],
        default_integrations=False,
        trace_lifecycle="stream",
    )
    items = capture_items("span", "event")

    schema = Schema(query=Query)

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server_sync():
        data = request.get_json()
        result = schema.execute(data["query"], operation_name=data.get("operationName"))
        return jsonify(result.data), 200

    query = {
        "query": "query ErrorQuery { goodbye }",
        "operationName": "ErrorQuery",
    }
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    sentry_sdk.get_client().flush()

    events = [item.payload for item in items if item.type == "event"]
    assert len(events) == 1

    (event,) = events
    assert len(event["breadcrumbs"]) == 1

    breadcrumbs = event["breadcrumbs"]["values"]
    assert len(breadcrumbs) == 1

    (breadcrumb,) = breadcrumbs
    assert breadcrumb["category"] == "graphql.operation"
    assert breadcrumb["data"]["operation_name"] == query["operationName"]
    assert breadcrumb["data"]["operation_type"] == "query"
    assert breadcrumb["type"] == "default"



from __future__ import annotations

import functools
from typing import Any

import pytest
from litestar import Controller, Litestar, get
from litestar.exceptions import HTTPException
from litestar.logging.config import LoggingConfig
from litestar.middleware import AbstractMiddleware
from litestar.middleware.logging import LoggingMiddlewareConfig
from litestar.middleware.rate_limit import RateLimitConfig
from litestar.middleware.session.server_side import ServerSideSessionConfig
from litestar.testing import TestClient

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.integrations.litestar import LitestarIntegration
from tests.conftest import ApproxDict
from tests.integrations.conftest import parametrize_test_configurable_status_codes


def litestar_app_factory(middleware=None, debug=True, exception_handlers=None):
    class MyController(Controller):
        path = "/controller"

        @get("/error")
        async def controller_error(self) -> None:
            raise Exception("Whoa")

    @get("/some_url")
    async def homepage_handler() -> "dict[str, Any]":
        1 / 0
        return {"status": "ok"}

    @get("/custom_error", name="custom_name")
    async def custom_error() -> Any:
        raise Exception("Too Hot")

    @get("/message")
    async def message() -> "dict[str, Any]":
        capture_message("hi")
        return {"status": "ok"}

    @get("/message/{message_id:str}")
    async def message_with_id() -> "dict[str, Any]":
        capture_message("hi")
        return {"status": "ok"}

    logging_config = LoggingConfig()

    app = Litestar(
        route_handlers=[
            homepage_handler,
            custom_error,
            message,
            message_with_id,
            MyController,
        ],
        debug=debug,
        middleware=middleware,
        logging_config=logging_config,
        exception_handlers=exception_handlers,
    )

    return app


@pytest.mark.parametrize(
    "test_url,expected_error,expected_message,expected_tx_name",
    [
        (
            "/some_url",
            ZeroDivisionError,
            "division by zero",
            "tests.integrations.litestar.test_litestar.litestar_app_factory.<locals>.homepage_handler",
        ),
        (
            "/custom_error",
            Exception,
            "Too Hot",
            "custom_name",
        ),
        (
            "/controller/error",
            Exception,
            "Whoa",
            "tests.integrations.litestar.test_litestar.litestar_app_factory.<locals>.MyController.controller_error",
        ),
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_catch_exceptions(
    sentry_init,
    capture_exceptions,
    capture_events,
    capture_items,
    test_url,
    expected_error,
    expected_message,
    expected_tx_name,
    span_streaming,
):
    sentry_init(
        integrations=[LitestarIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )
    litestar_app = litestar_app_factory()
    client = TestClient(litestar_app)
    exceptions = capture_exceptions()
    if span_streaming:
        items = capture_items("event")

        try:
            client.get(test_url)
        except Exception:
            pass

        (exc,) = exceptions
        assert isinstance(exc, expected_error)
        assert str(exc) == expected_message

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        try:
            client.get(test_url)
        except Exception:
            pass

        (exc,) = exceptions
        assert isinstance(exc, expected_error)
        assert str(exc) == expected_message

        (event,) = events
    assert expected_tx_name in event["transaction"]
    assert event["exception"]["values"][0]["mechanism"]["type"] == "litestar"


@pytest.mark.parametrize(
    "test_url,expected_tx_name",
    [
        (
            "/some_url",
            "tests.integrations.litestar.test_litestar.litestar_app_factory.<locals>.homepage_handler",
        ),
        (
            "/custom_error",
            "custom_name",
        ),
        (
            "/controller/error",
            "tests.integrations.litestar.test_litestar.litestar_app_factory.<locals>.MyController.controller_error",
        ),
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_name_and_source(
    sentry_init,
    capture_events,
    test_url,
    expected_tx_name,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[LitestarIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    litestar_app = litestar_app_factory()
    client = TestClient(litestar_app)

    if span_streaming:
        items = capture_items("span")

        try:
            client.get(test_url)
        except Exception:
            pass

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        spans = [span for span in spans if expected_tx_name in span["name"]]
        assert len(spans) == 1
        assert spans[0]["attributes"]["sentry.segment.name.source"] == "component"
    else:
        events = capture_events()

        try:
            client.get(test_url)
        except Exception:
            pass

        (_, transaction) = events
        assert expected_tx_name in transaction["transaction"]
        assert transaction["transaction_info"] == {"source": "component"}


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_spans(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[LitestarIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    logging_config = LoggingMiddlewareConfig()
    session_config = ServerSideSessionConfig()
    rate_limit_config = RateLimitConfig(rate_limit=("hour", 5))

    litestar_app = litestar_app_factory(
        middleware=[
            session_config.middleware,
            logging_config.middleware,
            rate_limit_config.middleware,
        ]
    )
    client = TestClient(
        litestar_app, raise_server_exceptions=False, base_url="http://testserver.local"
    )
    if span_streaming:
        items = capture_items("span")

        client.get("/message")

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        expected = {"SessionMiddleware", "LoggingMiddleware", "RateLimitMiddleware"}
        found = set()

        litestar_spans = (
            span
            for span in spans
            if span["attributes"]["sentry.op"] == "middleware.litestar"
        )

        for span in litestar_spans:
            assert span["name"] in expected
            assert span["name"] not in found
            found.add(span["name"])
            assert span["name"] == span["attributes"]["middleware.name"]
    else:
        events = capture_events()

        client.get("/message")

        (_, transaction_event) = events

        expected = {"SessionMiddleware", "LoggingMiddleware", "RateLimitMiddleware"}
        found = set()

        litestar_spans = (
            span
            for span in transaction_event["spans"]
            if span["op"] == "middleware.litestar"
        )

        for span in litestar_spans:
            assert span["description"] in expected
            assert span["description"] not in found
            found.add(span["description"])
            assert span["description"] == span["tags"]["litestar.middleware_name"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_callback_spans(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    class SampleMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send) -> None:
            async def do_stuff(message):
                if message["type"] == "http.response.start":
                    # do something here.
                    pass
                await send(message)

            await self.app(scope, receive, do_stuff)

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[LitestarIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    litestar_app = litestar_app_factory(middleware=[SampleMiddleware])
    client = TestClient(litestar_app, raise_server_exceptions=False)
    if span_streaming:
        items = capture_items("span")

        client.get("/message")

        spans = [item.payload for item in items]

        expected_litestar_spans = [
            {
                "name": "SampleMiddleware",
                "attributes": ApproxDict(
                    {
                        "middleware.name": "SampleMiddleware",
                        "sentry.op": "middleware.litestar",
                    },
                ),
            },
            {
                "name": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
                "attributes": ApproxDict(
                    {
                        "middleware.name": "SampleMiddleware",
                        "sentry.op": "middleware.litestar.send",
                    }
                ),
            },
            {
                "name": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
                "attributes": ApproxDict(
                    {
                        "middleware.name": "SampleMiddleware",
                        "sentry.op": "middleware.litestar.send",
                    }
                ),
            },
        ]

        def is_matching_span(expected_span, actual_span):
            return (
                expected_span["name"] == actual_span["name"]
                and expected_span["attributes"] == actual_span["attributes"]
            )

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        actual_litestar_spans = list(
            span
            for span in spans
            if "middleware.litestar" in span["attributes"].get("sentry.op")
        )
    else:
        events = capture_events()

        client.get("/message")

        (_, transaction_events) = events

        expected_litestar_spans = [
            {
                "op": "middleware.litestar",
                "description": "SampleMiddleware",
                "tags": {"litestar.middleware_name": "SampleMiddleware"},
            },
            {
                "op": "middleware.litestar.send",
                "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
                "tags": {"litestar.middleware_name": "SampleMiddleware"},
            },
            {
                "op": "middleware.litestar.send",
                "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
                "tags": {"litestar.middleware_name": "SampleMiddleware"},
            },
        ]

        def is_matching_span(expected_span, actual_span):
            return (
                expected_span["op"] == actual_span["op"]
                and expected_span["description"] == actual_span["description"]
                and expected_span["tags"] == actual_span["tags"]
            )

        actual_litestar_spans = list(
            span
            for span in transaction_events["spans"]
            if "middleware.litestar" in span["op"]
        )

    assert len(actual_litestar_spans) == 3

    for expected_span in expected_litestar_spans:
        assert any(
            is_matching_span(expected_span, actual_span)
            for actual_span in actual_litestar_spans
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_receive_send(sentry_init, capture_items, span_streaming):
    class SampleReceiveSendMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send):
            message = await receive()
            assert message
            assert message["type"] == "http.request"

            send_output = await send({"type": "something-unimportant"})
            assert send_output is None

            await self.app(scope, receive, send)

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[LitestarIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )
    litestar_app = litestar_app_factory(middleware=[SampleReceiveSendMiddleware])

    client = TestClient(litestar_app, raise_server_exceptions=False)
    # See SampleReceiveSendMiddleware.__call__ above for assertions of correct behavior
    client.get("/message")


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_partial_receive_send(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    class SamplePartialReceiveSendMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send):
            message = await receive()
            assert message
            assert message["type"] == "http.request"

            send_output = await send({"type": "something-unimportant"})
            assert send_output is None

            async def my_receive(*args, **kwargs):
                pass

            async def my_send(*args, **kwargs):
                pass

            partial_receive = functools.partial(my_receive)
            partial_send = functools.partial(my_send)

            await self.app(scope, partial_receive, partial_send)

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[LitestarIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    litestar_app = litestar_app_factory(middleware=[SamplePartialReceiveSendMiddleware])
    client = TestClient(litestar_app, raise_server_exceptions=False)

    if span_streaming:
        items = capture_items("span")

        # See SamplePartialReceiveSendMiddleware.__call__ above for assertions of correct behavior
        client.get("/message")

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        expected_litestar_spans = [
            {
                "name": "SamplePartialReceiveSendMiddleware",
                "attributes": ApproxDict(
                    {
                        "middleware.name": "SamplePartialReceiveSendMiddleware",
                        "sentry.op": "middleware.litestar",
                    }
                ),
            },
            {
                "name": "TestClientTransport.create_receive.<locals>.receive",
                "attributes": ApproxDict(
                    {
                        "middleware.name": "SamplePartialReceiveSendMiddleware",
                        "sentry.op": "middleware.litestar.receive",
                    }
                ),
            },
            {
                "name": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
                "attributes": ApproxDict(
                    {
                        "middleware.name": "SamplePartialReceiveSendMiddleware",
                        "sentry.op": "middleware.litestar.send",
                    }
                ),
            },
        ]

        def is_matching_span(expected_span, actual_span):
            return (
                actual_span["name"].startswith(expected_span["name"])
                and expected_span["attributes"] == actual_span["attributes"]
            )

        actual_litestar_spans = list(
            span
            for span in spans
            if "middleware.litestar" in span["attributes"].get("sentry.op")
        )
    else:
        events = capture_events()

        # See SamplePartialReceiveSendMiddleware.__call__ above for assertions of correct behavior
        client.get("/message")

        (_, transaction_events) = events

        expected_litestar_spans = [
            {
                "op": "middleware.litestar",
                "description": "SamplePartialReceiveSendMiddleware",
                "tags": {
                    "litestar.middleware_name": "SamplePartialReceiveSendMiddleware"
                },
            },
            {
                "op": "middleware.litestar.receive",
                "description": "TestClientTransport.create_receive.<locals>.receive",
                "tags": {
                    "litestar.middleware_name": "SamplePartialReceiveSendMiddleware"
                },
            },
            {
                "op": "middleware.litestar.send",
                "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
                "tags": {
                    "litestar.middleware_name": "SamplePartialReceiveSendMiddleware"
                },
            },
        ]

        def is_matching_span(expected_span, actual_span):
            return (
                expected_span["op"] == actual_span["op"]
                and actual_span["description"].startswith(expected_span["description"])
                and expected_span["tags"] == actual_span["tags"]
            )

        actual_litestar_spans = list(
            span
            for span in transaction_events["spans"]
            if "middleware.litestar" in span["op"]
        )
    assert len(actual_litestar_spans) == 3

    for expected_span in expected_litestar_spans:
        assert any(
            is_matching_span(expected_span, actual_span)
            for actual_span in actual_litestar_spans
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[LitestarIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    logging_config = LoggingMiddlewareConfig()
    session_config = ServerSideSessionConfig()
    rate_limit_config = RateLimitConfig(rate_limit=("hour", 5))

    litestar_app = litestar_app_factory(
        middleware=[
            session_config.middleware,
            logging_config.middleware,
            rate_limit_config.middleware,
        ]
    )
    client = TestClient(
        litestar_app, raise_server_exceptions=False, base_url="http://testserver.local"
    )
    if span_streaming:
        items = capture_items("span")

        client.get("/message")

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        for span in spans:
            if span["attributes"]["sentry.origin"] == "auto.http.httpx":
                continue
            assert span["attributes"]["sentry.origin"] == "auto.http.litestar"
    else:
        events = capture_events()

        client.get("/message")

        (_, event) = events

        assert event["contexts"]["trace"]["origin"] == "auto.http.litestar"
        for span in event["spans"]:
            assert span["origin"] == "auto.http.litestar"


@pytest.mark.parametrize(
    "is_send_default_pii",
    [
        True,
        False,
    ],
    ids=[
        "send_default_pii=True",
        "send_default_pii=False",
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_litestar_scope_user_on_exception_event(
    sentry_init,
    capture_exceptions,
    capture_events,
    capture_items,
    is_send_default_pii,
    span_streaming,
):
    class TestUserMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send):
            scope["user"] = {
                "email": "lennon@thebeatles.com",
                "username": "john",
                "id": "1",
            }
            await self.app(scope, receive, send)

    sentry_init(
        integrations=[LitestarIntegration()],
        send_default_pii=is_send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    litestar_app = litestar_app_factory(middleware=[TestUserMiddleware])
    client = TestClient(litestar_app)
    exceptions = capture_exceptions()
    if span_streaming:
        items = capture_items("event")

        # This request intentionally raises an exception
        try:
            client.get("/some_url")
        except Exception:
            pass

        assert len(exceptions) == 1
        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        # This request intentionally raises an exception
        try:
            client.get("/some_url")
        except Exception:
            pass

        assert len(exceptions) == 1
        assert len(events) == 1
        (event,) = events

    if is_send_default_pii:
        assert "user" in event
        assert event["user"] == {
            "email": "lennon@thebeatles.com",
            "username": "john",
            "id": "1",
        }
    else:
        assert "user" not in event


@parametrize_test_configurable_status_codes
@pytest.mark.parametrize("span_streaming", [True, False])
def test_configurable_status_codes_handler(
    sentry_init,
    capture_events,
    capture_items,
    failed_request_status_codes,
    status_code,
    expected_error,
    span_streaming,
):
    integration_kwargs = (
        {"failed_request_status_codes": failed_request_status_codes}
        if failed_request_status_codes is not None
        else {}
    )
    sentry_init(
        integrations=[LitestarIntegration(**integration_kwargs)],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @get("/error")
    async def error() -> None:
        raise HTTPException(status_code=status_code)

    app = Litestar([error])
    client = TestClient(app)

    if span_streaming:
        items = capture_items("event")

        client.get("/error")

        events = [item.payload for item in items]
    else:
        events = capture_events()

        client.get("/error")

    assert len(events) == int(expected_error)


@parametrize_test_configurable_status_codes
@pytest.mark.parametrize("span_streaming", [True, False])
def test_configurable_status_codes_middleware(
    sentry_init,
    capture_events,
    capture_items,
    failed_request_status_codes,
    status_code,
    expected_error,
    span_streaming,
):
    integration_kwargs = (
        {"failed_request_status_codes": failed_request_status_codes}
        if failed_request_status_codes is not None
        else {}
    )

    sentry_init(
        integrations=[LitestarIntegration(**integration_kwargs)],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    def create_raising_middleware(app):
        async def raising_middleware(scope, receive, send):
            raise HTTPException(status_code=status_code)

        return raising_middleware

    @get("/error")
    async def error() -> None: ...

    app = Litestar([error], middleware=[create_raising_middleware])
    client = TestClient(app)

    if span_streaming:
        items = capture_items("event")

        client.get("/error")

        events = [item.payload for item in items]
    else:
        events = capture_events()

        client.get("/error")

    assert len(events) == int(expected_error)


@pytest.mark.parametrize("span_streaming", [True, False])
def test_catch_non_http_exceptions_in_middleware(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[LitestarIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    def create_raising_middleware(app):
        async def raising_middleware(scope, receive, send):
            raise RuntimeError("Too Hot")

        return raising_middleware

    @get("/error")
    async def error() -> None: ...

    app = Litestar([error], middleware=[create_raising_middleware])
    client = TestClient(app)

    if span_streaming:
        items = capture_items("event")

        try:
            client.get("/error")
        except RuntimeError:
            pass

        events = [item.payload for item in items]
    else:
        events = capture_events()

        try:
            client.get("/error")
        except RuntimeError:
            pass

    assert len(events) == 1
    event_exception = events[0]["exception"]["values"][0]
    assert event_exception["type"] == "RuntimeError"
    assert event_exception["value"] == "Too Hot"



import json
from unittest import mock  # python 3.3 and above

import httpx
import pytest
from cohere import ChatMessage, Client
from httpx import Client as HTTPXClient

import sentry_sdk
from sentry_sdk import start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.cohere import CohereIntegration


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize(
    "send_default_pii, include_prompts",
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_nonstreaming_chat(
    sentry_init,
    capture_events,
    capture_items,
    send_default_pii,
    include_prompts,
    span_streaming,
):
    sentry_init(
        integrations=[CohereIntegration(include_prompts=include_prompts)],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    client = Client(api_key="z")
    HTTPXClient.request = mock.Mock(
        return_value=httpx.Response(
            200,
            json={
                "text": "the model response",
                "meta": {
                    "billed_units": {
                        "output_tokens": 10,
                        "input_tokens": 20,
                    }
                },
            },
        )
    )

    if span_streaming:
        items = capture_items("span")

        with start_transaction(name="cohere tx"):
            response = client.chat(
                model="some-model",
                chat_history=[ChatMessage(role="SYSTEM", message="some context")],
                message="hello",
            ).text

        assert response == "the model response"
        sentry_sdk.flush()

        assert len(items) == 1
        span = items[0].payload

        assert span["attributes"]["sentry.op"] == "ai.chat_completions.create.cohere"
        assert span["attributes"][SPANDATA.AI_MODEL_ID] == "some-model"

        if send_default_pii and include_prompts:
            assert (
                '{"role": "system", "content": "some context"}'
                in span["attributes"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert (
                '{"role": "user", "content": "hello"}'
                in span["attributes"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert "the model response" in span["attributes"][SPANDATA.AI_RESPONSES]
        else:
            assert SPANDATA.AI_INPUT_MESSAGES not in span["attributes"]
            assert SPANDATA.AI_RESPONSES not in span["attributes"]

        assert span["attributes"]["gen_ai.usage.output_tokens"] == 10
        assert span["attributes"]["gen_ai.usage.input_tokens"] == 20
        assert span["attributes"]["gen_ai.usage.total_tokens"] == 30
    else:
        events = capture_events()

        with start_transaction(name="cohere tx"):
            response = client.chat(
                model="some-model",
                chat_history=[ChatMessage(role="SYSTEM", message="some context")],
                message="hello",
            ).text

        assert response == "the model response"
        tx = events[0]
        assert tx["type"] == "transaction"
        assert len(tx["spans"]) == 1
        span = tx["spans"][0]
        assert span["op"] == "ai.chat_completions.create.cohere"
        assert span["data"][SPANDATA.AI_MODEL_ID] == "some-model"

        if send_default_pii and include_prompts:
            assert (
                '{"role": "system", "content": "some context"}'
                in span["data"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert (
                '{"role": "user", "content": "hello"}'
                in span["data"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert "the model response" in span["data"][SPANDATA.AI_RESPONSES]
        else:
            assert SPANDATA.AI_INPUT_MESSAGES not in span["data"]
            assert SPANDATA.AI_RESPONSES not in span["data"]

        assert span["data"]["gen_ai.usage.output_tokens"] == 10
        assert span["data"]["gen_ai.usage.input_tokens"] == 20
        assert span["data"]["gen_ai.usage.total_tokens"] == 30


# noinspection PyTypeChecker
@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize(
    "send_default_pii, include_prompts",
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_streaming_chat(
    sentry_init,
    capture_events,
    capture_items,
    send_default_pii,
    include_prompts,
    span_streaming,
):
    sentry_init(
        integrations=[CohereIntegration(include_prompts=include_prompts)],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    client = Client(api_key="z")
    HTTPXClient.send = mock.Mock(
        return_value=httpx.Response(
            200,
            content="\n".join(
                [
                    json.dumps({"event_type": "text-generation", "text": "the model "}),
                    json.dumps({"event_type": "text-generation", "text": "response"}),
                    json.dumps(
                        {
                            "event_type": "stream-end",
                            "finish_reason": "COMPLETE",
                            "response": {
                                "text": "the model response",
                                "meta": {
                                    "billed_units": {
                                        "output_tokens": 10,
                                        "input_tokens": 20,
                                    }
                                },
                            },
                        }
                    ),
                ]
            ),
        )
    )

    if span_streaming:
        items = capture_items("span")

        with start_transaction(name="cohere tx"):
            responses = list(
                client.chat_stream(
                    model="some-model",
                    chat_history=[ChatMessage(role="SYSTEM", message="some context")],
                    message="hello",
                )
            )
            response_string = responses[-1].response.text

        assert response_string == "the model response"
        sentry_sdk.flush()

        assert len(items) == 1
        span = items[0].payload

        assert span["attributes"]["sentry.op"] == "ai.chat_completions.create.cohere"
        assert span["attributes"][SPANDATA.AI_MODEL_ID] == "some-model"

        if send_default_pii and include_prompts:
            assert (
                '{"role": "system", "content": "some context"}'
                in span["attributes"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert (
                '{"role": "user", "content": "hello"}'
                in span["attributes"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert "the model response" in span["attributes"][SPANDATA.AI_RESPONSES]
        else:
            assert SPANDATA.AI_INPUT_MESSAGES not in span["attributes"]
            assert SPANDATA.AI_RESPONSES not in span["attributes"]

        assert span["attributes"]["gen_ai.usage.output_tokens"] == 10
        assert span["attributes"]["gen_ai.usage.input_tokens"] == 20
        assert span["attributes"]["gen_ai.usage.total_tokens"] == 30
    else:
        events = capture_events()

        with start_transaction(name="cohere tx"):
            responses = list(
                client.chat_stream(
                    model="some-model",
                    chat_history=[ChatMessage(role="SYSTEM", message="some context")],
                    message="hello",
                )
            )
            response_string = responses[-1].response.text

        assert response_string == "the model response"
        tx = events[0]
        assert tx["type"] == "transaction"
        assert len(tx["spans"]) == 1
        span = tx["spans"][0]
        assert span["op"] == "ai.chat_completions.create.cohere"
        assert span["data"][SPANDATA.AI_MODEL_ID] == "some-model"

        if send_default_pii and include_prompts:
            assert (
                '{"role": "system", "content": "some context"}'
                in span["data"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert (
                '{"role": "user", "content": "hello"}'
                in span["data"][SPANDATA.AI_INPUT_MESSAGES]
            )
            assert "the model response" in span["data"][SPANDATA.AI_RESPONSES]
        else:
            assert SPANDATA.AI_INPUT_MESSAGES not in span["data"]
            assert SPANDATA.AI_RESPONSES not in span["data"]

        assert span["data"]["gen_ai.usage.output_tokens"] == 10
        assert span["data"]["gen_ai.usage.input_tokens"] == 20
        assert span["data"]["gen_ai.usage.total_tokens"] == 30


def test_bad_chat(sentry_init, capture_events):
    sentry_init(integrations=[CohereIntegration()], traces_sample_rate=1.0)
    events = capture_events()

    client = Client(api_key="z")
    HTTPXClient.request = mock.Mock(
        side_effect=httpx.HTTPError("API rate limit reached")
    )
    with pytest.raises(httpx.HTTPError):
        client.chat(model="some-model", message="hello")

    (event, transaction) = events
    assert event["level"] == "error"
    assert transaction["contexts"]["trace"]["status"] == "internal_error"


def test_span_status_error(sentry_init, capture_events):
    sentry_init(integrations=[CohereIntegration()], traces_sample_rate=1.0)
    events = capture_events()

    with start_transaction(name="test"):
        client = Client(api_key="z")
        HTTPXClient.request = mock.Mock(
            side_effect=httpx.HTTPError("API rate limit reached")
        )
        with pytest.raises(httpx.HTTPError):
            client.chat(model="some-model", message="hello")

    (error, transaction) = events
    assert error["level"] == "error"
    assert transaction["spans"][0]["status"] == "internal_error"
    assert transaction["spans"][0]["tags"]["status"] == "internal_error"


def test_span_status_error_streaming(sentry_init, capture_events, capture_items):
    sentry_init(
        integrations=[CohereIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    client = Client(api_key="z")
    HTTPXClient.request = mock.Mock(
        side_effect=httpx.HTTPError("API rate limit reached")
    )
    with start_transaction(name="test"):
        with pytest.raises(httpx.HTTPError):
            client.chat(model="some-model", message="hello")

    sentry_sdk.flush()

    assert len(items) == 1
    span = items[0].payload
    assert span["status"] == "error"


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize(
    "send_default_pii, include_prompts",
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_embed(
    sentry_init,
    capture_events,
    capture_items,
    send_default_pii,
    include_prompts,
    span_streaming,
):
    sentry_init(
        integrations=[CohereIntegration(include_prompts=include_prompts)],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    client = Client(api_key="z")
    HTTPXClient.request = mock.Mock(
        return_value=httpx.Response(
            200,
            json={
                "response_type": "embeddings_floats",
                "id": "1",
                "texts": ["hello"],
                "embeddings": [[1.0, 2.0, 3.0]],
                "meta": {
                    "billed_units": {
                        "input_tokens": 10,
                    }
                },
            },
        )
    )

    if span_streaming:
        items = capture_items("span")

        with start_transaction(name="cohere tx"):
            response = client.embed(texts=["hello"], model="text-embedding-3-large")

        assert len(response.embeddings[0]) == 3
        sentry_sdk.flush()

        assert len(items) == 1
        span = items[0].payload

        assert span["attributes"]["sentry.op"] == "ai.embeddings.create.cohere"
        if send_default_pii and include_prompts:
            assert "hello" in span["attributes"][SPANDATA.AI_INPUT_MESSAGES]
        else:
            assert SPANDATA.AI_INPUT_MESSAGES not in span["attributes"]

        assert span["attributes"]["gen_ai.usage.input_tokens"] == 10
        assert span["attributes"]["gen_ai.usage.total_tokens"] == 10
    else:
        events = capture_events()

        with start_transaction(name="cohere tx"):
            response = client.embed(texts=["hello"], model="text-embedding-3-large")

        assert len(response.embeddings[0]) == 3

        tx = events[0]
        assert tx["type"] == "transaction"
        assert len(tx["spans"]) == 1
        span = tx["spans"][0]
        assert span["op"] == "ai.embeddings.create.cohere"
        if send_default_pii and include_prompts:
            assert "hello" in span["data"][SPANDATA.AI_INPUT_MESSAGES]
        else:
            assert SPANDATA.AI_INPUT_MESSAGES not in span["data"]

        assert span["data"]["gen_ai.usage.input_tokens"] == 10
        assert span["data"]["gen_ai.usage.total_tokens"] == 10


def test_span_origin_chat(sentry_init, capture_events):
    sentry_init(
        integrations=[CohereIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    client = Client(api_key="z")
    HTTPXClient.request = mock.Mock(
        return_value=httpx.Response(
            200,
            json={
                "text": "the model response",
                "meta": {
                    "billed_units": {
                        "output_tokens": 10,
                        "input_tokens": 20,
                    }
                },
            },
        )
    )

    with start_transaction(name="cohere tx"):
        client.chat(
            model="some-model",
            chat_history=[ChatMessage(role="SYSTEM", message="some context")],
            message="hello",
        ).text

    (event,) = events

    assert event["contexts"]["trace"]["origin"] == "manual"
    assert event["spans"][0]["origin"] == "auto.ai.cohere"


def test_span_origin_embed(sentry_init, capture_events):
    sentry_init(
        integrations=[CohereIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    client = Client(api_key="z")
    HTTPXClient.request = mock.Mock(
        return_value=httpx.Response(
            200,
            json={
                "response_type": "embeddings_floats",
                "id": "1",
                "texts": ["hello"],
                "embeddings": [[1.0, 2.0, 3.0]],
                "meta": {
                    "billed_units": {
                        "input_tokens": 10,
                    }
                },
            },
        )
    )

    with start_transaction(name="cohere tx"):
        client.embed(texts=["hello"], model="text-embedding-3-large")

    (event,) = events

    assert event["contexts"]["trace"]["origin"] == "manual"
    assert event["spans"][0]["origin"] == "auto.ai.cohere"
